from collections import Counter
from copy import deepcopy
from math import floor
from typing import List, Dict, Self, Optional

from sklearn.mixture import GaussianMixture

import pandas as pd
import numpy as np

from kymata.ippm.constants import LATENCY_OFFSET, ANOMALOUS_TAG

class CustomClusterer:
    """
        You need to override these methods to create a new clusterer. self.labels_ assigns each datapoint
        in df to a cluster. Set anomalies to ANOMALOUS_TAG.
    """

    def __init__(self):
        self.labels_ = []

    def fit(self, df: pd.DataFrame) -> Self:
        return self


class MaxPooler(CustomClusterer):
    def __init__(self, label_significance_threshold: int = 15, label_size: int = 25):
        super().__init__()
        self._label_significance_threshold = label_significance_threshold
        self._label_size = label_size

    def fit(self, df: pd.DataFrame) -> Self:
        self.labels_ = self._assign_points_to_labels(df)
        count_of_data_per_label = dict(Counter(self.labels_))
        self.labels_ = self._tag_labels_below_label_significance_threshold_as_anomalies(count_of_data_per_label)
        return self

    def _assign_points_to_labels(self, df_with_latency: pd.DataFrame, latency_col_index: int = 0) -> List[int]:
        def __get_bin_index(latency: float) -> int:
            return floor((latency + LATENCY_OFFSET) / self._label_size)  # floor because end is exclusive

        return list(map(__get_bin_index, df_with_latency.iloc[:, latency_col_index]))

    def _tag_labels_below_label_significance_threshold_as_anomalies(
            self, count_of_data_per_bin: Dict[int, int]
    ) -> List[int]:
        for label, count in count_of_data_per_bin.items():
            if count < self._label_significance_threshold:
                self.labels_ = self._map_label_to_new_label(label, ANOMALOUS_TAG)
        return self.labels_

    def _map_label_to_new_label(self, old_label: int, new_label: int) -> List[int]:
        return list(map(
            lambda x: new_label if x == old_label else x,
            self.labels_
        ))


class AdaptiveMaxPooler(MaxPooler):
    def __init__(self, label_significance_threshold: int = 5, base_label_size: int = 10):
        super().__init__(label_significance_threshold, base_label_size)
        self.labels_ = None
        self._base_label_size = base_label_size

    def fit(self, df: pd.DataFrame) -> Self:
        self.labels_ = super()._assign_points_to_labels(df)
        count_of_data_per_label = dict(Counter(self.labels_))
        self.labels_ = super()._tag_labels_below_label_significance_threshold_as_anomalies(count_of_data_per_label)
        self.labels_ = self._merge_significant_labels(df)
        return self

    def _merge_significant_labels(self, df: pd.DataFrame) -> List[int]:
        def __is_insignificant_label(end_index: int) -> bool:
            return self.labels_[end_index] == ANOMALOUS_TAG

        def __labels_are_not_adjacent_and_previous_was_significant(end_index: int) -> bool:
            if __is_insignificant_label(end_index - 1):
                # we do not care about the gap between insignificant label and significant label
                return False
            current_latency = df.iloc[end_index, 0]
            previous_latency = df.iloc[end_index - 1, 0]
            return current_latency - previous_latency > self._base_label_size

        def __reached_end(end_index: int) -> bool:
            return end_index == len(self.labels_) - 1

        new_label = 0
        start_index = 0
        for end_index in range(len(self.labels_)):
            if (
                    __is_insignificant_label(end_index)
                    or __labels_are_not_adjacent_and_previous_was_significant(end_index)
                    or __reached_end(end_index)
            ):
                self.labels_ = self._merge_labels(
                    start_index,
                    end_index,
                    new_label,
                    __is_insignificant_label(end_index),
                )
                new_label += 1

                if __is_insignificant_label(end_index):
                    start_index = end_index + 1
                else:
                    start_index = end_index

        return self.labels_

    def _merge_labels(self, start_index: int, end_index: int, new_label: int, current_label_is_anomalous: bool):
        if start_index < end_index:
            if not current_label_is_anomalous:
                # inclusive merge
                end_index += 1

            for index in range(start_index, end_index):
                self.labels_[index] = new_label

        return self.labels_


class CustomGMM(CustomClusterer):
    def __init__(
            self,
            number_of_clusters_upper_bound: int = 5,
            covariance_type: str = 'full',
            max_iter: int = 1000,
            n_init: int = 8,
            init_params: str = 'kmeans',
            random_state: Optional[int] = None,
            should_evaluate_using_AIC: bool = False
    ):
        """
        Akaike Information Criterion: -2 * log(L) + 2 * k where L = likelihood
                                                                k = # of model parameters
        Bayesian Information Criterion: -2 * log(L) + log(N) * where L and k defined as above
                                                                     N = # of datapoints

        :param number_of_clusters_upper_bound:
        :param covariance_type:
        :param max_iter:
        :param n_init:
        :param init_params:
        :param random_state:
        :param should_evaluate_using_AIC:
        """
        super().__init__()
        self._number_of_clusters_upper_bound = number_of_clusters_upper_bound
        self._covariance_type = covariance_type
        self._max_iter = max_iter
        self._n_init = n_init
        self._init_params = init_params
        self._random_state = random_state
        self._should_evaluate_using_AIC = should_evaluate_using_AIC

    def fit(self, df: pd.DataFrame) -> Self:
        optimal_model = self._grid_search_for_optimal_number_of_clusters(df)
        if optimal_model is not None:
            # None if no data.
            self.labels_ = optimal_model.predict(df)
            self.labels_ = self._tag_low_loglikelihood_points_as_anomalous(df, optimal_model)
        return self

    def _grid_search_for_optimal_number_of_clusters(self, df: pd.DataFrame) -> GaussianMixture:
        """
        Quick 101 to model evaluation:
            - Likelihood for a datapoint represents the probability that the datapoint came from the estimated distribution.
              Therefore, the higher the likelihood, the better the fit.
            - Log-Likelihood maps the Likelihood to (-inf, 0]. We still want to maximise this.
            - AIC and BIC use negative Log-Likeliood, so we now attempt to minimise them.
              You can interpret both of these metrics as negative log-likelihood with a model complexity penalty.

        :param df:
        :return:
        """
        def __evaluate_fit(data: pd.DataFrame, fitted_model: GaussianMixture) -> float:
            return fitted_model.aic(data) if self._should_evaluate_using_AIC else fitted_model.bic(data)

        optimal_penalised_loglikelihood = np.inf
        optimal_model = None
        for number_of_clusters in range(1, self._number_of_clusters_upper_bound):
            if number_of_clusters > len(df):
                break

            copy_of_df = deepcopy(df)

            model = GaussianMixture(
                n_components=number_of_clusters,
                covariance_type=self._covariance_type,
                max_iter=self._max_iter,
                n_init=self._n_init,
                init_params=self._init_params,
                random_state=self._random_state
            )

            model.fit(copy_of_df)

            penalised_loglikelihood = __evaluate_fit(copy_of_df, model)
            if penalised_loglikelihood < optimal_penalised_loglikelihood:
                optimal_penalised_loglikelihood = penalised_loglikelihood
                optimal_model = model

        return optimal_model

    def _tag_low_loglikelihood_points_as_anomalous(
            self,
            df: pd.DataFrame,
            optimal_model: GaussianMixture,
            anomaly_percentile_threshold: int = 5
    ) -> List[int]:
        def __update_labels_to_anomalous_label(log_likelihoods, anomaly_threshold):
            # we do > cus more negative loglikelihood, the higher the likelihood.
            return list(map(
                lambda x: ANOMALOUS_TAG if x[0] < anomaly_threshold else x[1],
                zip(log_likelihoods, self.labels_)
            ))

        log_likelihood = optimal_model.score_samples(df)
        threshold = np.percentile(log_likelihood, anomaly_percentile_threshold)
        self.labels_ = __update_labels_to_anomalous_label(log_likelihood, threshold)
        return self.labels_




from os import getcwd, getenv
from pathlib import Path
from typing import Optional

from kymata.io.file import PathType


_DATA_PATH_ENVIRONMENT_VAR_NAME = "KYMATA_DATA_ROOT"
DATA_DIR_NAME = "kymata-core-data"

# Places downloaded datasets could go, in order of preference
_preferred_default_data_locations = [
    Path(Path(__file__).parent.parent.parent),  # kymata/../data_dir (next to kymata dir)
    Path(getcwd()),                             # <cwd>/data_dir
    Path(Path.home(), "Documents"),             # ~/Documents/data_dir
    Path(Path.home()),                          # ~/data_dir
]


def data_root_path(data_root: Optional[PathType] = None) -> Path:

    # Check if the data root has been specified

    # Might be in an environmental variable
    if data_root is None:
        data_root: PathType | None = getenv(_DATA_PATH_ENVIRONMENT_VAR_NAME, default=None)

    # Might have been supplied as an argument
    if data_root is not None:
        if isinstance(data_root, str):
            data_root = Path(data_root)
        # Data root specified
        if not data_root.exists():
            raise FileNotFoundError(f"data_root {str(data_root)} specified but does not exist")
        if not data_root.is_dir():
            raise NotADirectoryError(f"Please specify a directory ({str(data_root)} is not a directory)")

        return data_root

    else:
        # Data root not specified

        # Check if the data root already exists
        for loc in _preferred_default_data_locations:
            if (here := Path(loc, DATA_DIR_NAME)).exists():
                data_root = here
                break

        # If not, attempt to create it
        if data_root is None:
            here: Path | None = None
            for loc in _preferred_default_data_locations:
                here = Path(loc, DATA_DIR_NAME)
                try:
                    here.mkdir()
                    break
                # If it fails for sensible reasons, no sweat, we'll fall through to the next option
                except (FileNotFoundError, OSError):
                    # Parent didn't exist, not writeable, etc
                    pass
            # Did we make it?
            if here is not None and here.exists():
                data_root = here
            else:
                raise FileNotFoundError("Failed to create data root directory")

        # Data root location has been derived, rather than prespecified, so feed that back to the user to avoid a
        # different location somehow being derived next time
        print(f"Data root set at {str(data_root)}.")
        print(f"Consider setting this as environmental variable {_DATA_PATH_ENVIRONMENT_VAR_NAME} to ensure it's reused"
              f" next time.")
        print(f"Hint: $> {_DATA_PATH_ENVIRONMENT_VAR_NAME}=\"{str(data_root)}\"")
        return data_root



from numpy import arange
from numpy.random import randint


def generate_derangement(n, mod=int(1e9)):  # approx 3ms runtime for n=400
    while True:
        v = arange(n)
        for j in range(n - 1, -1, -1):
            p = randint(0, j + 1)
            if v[p] % mod == j % mod:
                break
            else:
                v[j], v[p] = v[p], v[j]
        else:
            return v



from enum import StrEnum

from packaging import version
from io import TextIOWrapper
from pathlib import Path
from typing import Any
from warnings import warn
from zipfile import ZipFile, ZIP_LZMA

from numpy import frombuffer
from numpy.typing import NDArray
from sparse import COO

from kymata.entities.datatypes import HexelDType, LatencyDType, FunctionNameDType, SensorDType
from kymata.entities.expression import ExpressionSet, BLOCK_LEFT, BLOCK_RIGHT, BLOCK_SCALP, HexelExpressionSet, \
    SensorExpressionSet
from kymata.math.p_values import p_to_logp
from kymata.entities.sparse_data import expand_dims
from kymata.io.file import PathType, FileType, open_or_use


class _Keys(StrEnum):
    channels  = "channels"
    latencies = "latencies"
    functions = "functions"
    layers    = "layers"
    data      = "data"

    expressionset_type = "expressionset-type"


class _ExpressionSetTypeIdentifier(StrEnum):
    hexel  = "hexel"
    sensor = "sensor"

    @classmethod
    def from_expression_set(cls, expression_set: ExpressionSet):
        if isinstance(expression_set, HexelExpressionSet):
            return cls.hexel
        elif isinstance(expression_set, SensorExpressionSet):
            return cls.sensor
        else:
            raise NotImplementedError()

    def block_names(self) -> list[str]:
        if self == self.hexel:
            return [BLOCK_LEFT, BLOCK_RIGHT]
        elif self == self.sensor:
            return [BLOCK_SCALP]
        else:
            raise NotImplementedError()


# Format versioning of the saved file. Used to (eventually) ensure old saved files can always be loaded.
#
# - Increment MINOR versions when creating a breaking change, but only when we commit to indefinitely supporting
#   the new version.
# - Increment MAJOR version when creating a breaking change which makes the format fundamentally incompatible
#   with previous versions.
#
# All distinct versions should be documented. See `nkg_compatibility.py` for changes.
#
# This value should be saved into file called /_metadata/format-version.txt within an archive.
CURRENT_VERSION = "0.4"


def file_version(from_path_or_file: PathType | FileType) -> version.Version:
    with open_or_use(from_path_or_file, mode="rb") as archive, ZipFile(archive, "r") as zf:
        with TextIOWrapper(zf.open("_metadata/format-version.txt"), encoding="utf-8") as f:
            return version.parse(str(f.read()).strip())


def load_expression_set(from_path_or_file: PathType | FileType) -> ExpressionSet:
    """
    Loads an ExpressionSet from the specified path or file.

    The function determines the type of ExpressionSet (HexelExpressionSet or SensorExpressionSet)
    based on the data loaded from the provided path or file. It then constructs and returns an
    instance of the appropriate ExpressionSet subclass.

    Args:
        from_path_or_file (PathType | FileType): The path or file from which to load the data.

    Returns:
        ExpressionSet: An instance of either HexelExpressionSet or SensorExpressionSet,
                           depending on the type identifier in the data.

    Raises:
        KeyError: If required keys are missing in the data dictionary.
        ValueError: If the type identifier is not recognized.
    """
    _v, data_dict = _load_data(from_path_or_file)

    type_identifier = data_dict[_Keys.expressionset_type]

    if type_identifier == _ExpressionSetTypeIdentifier.hexel:
        return HexelExpressionSet(
            functions=data_dict[_Keys.functions],
            hexels_lh=[HexelDType(c) for c in data_dict[_Keys.channels][BLOCK_LEFT]],
            hexels_rh=[HexelDType(c) for c in data_dict[_Keys.channels][BLOCK_RIGHT]],
            latencies=data_dict[_Keys.latencies],
            data_lh=[data_dict[_Keys.data][BLOCK_LEFT][:, :, i]
                     for i in range(len(data_dict[_Keys.functions]))],
            data_rh=[data_dict[_Keys.data][BLOCK_RIGHT][:, :, i]
                     for i in range(len(data_dict[_Keys.functions]))],
        )
    elif type_identifier == _ExpressionSetTypeIdentifier.sensor:
        return SensorExpressionSet(
            functions=data_dict[_Keys.functions],
            sensors=[SensorDType(c) for c in data_dict[_Keys.channels][BLOCK_SCALP]],
            latencies=data_dict[_Keys.latencies],
            data=[data_dict[_Keys.data][BLOCK_SCALP][:, :, i]
                  for i in range(len(data_dict[_Keys.functions]))],
        )


def save_expression_set(expression_set: ExpressionSet,
                        to_path_or_file: PathType | FileType,
                        compression=ZIP_LZMA,
                        overwrite: bool = False):
    """
    Save the given ExpressionSet to a specified path or an already open file.

    This function saves the ExpressionSet data into a compressed file format.
    If a file path is provided, it creates and writes to the file. If an open file is supplied,
    it should be opened in "wb" mode. The overwrite flag is ignored if an open file is supplied.

    Args:
        expression_set (ExpressionSet): The ExpressionSet object to be saved.
        to_path_or_file (PathType | FileType): The path or open file where the ExpressionSet will be saved.
        compression: The compression method to use (default is ZIP_LZMA).
        overwrite (bool): If True, allows overwriting an existing file (default is False).

    Raises:
        FileExistsError: If the specified path already exists and overwrite is False.
        TypeError: If the provided path or file type is invalid.

    Notes:
        - The compression parameter should be compatible with the `ZipFile` class.
        - The function writes various metadata and data blocks in a structured format within the zip file.
    """

    if isinstance(to_path_or_file, str):
        to_path_or_file = Path(to_path_or_file)
    if isinstance(to_path_or_file, Path) and to_path_or_file.exists() and not overwrite:
        raise FileExistsError(to_path_or_file)

    with open_or_use(to_path_or_file, mode="wb") as f, ZipFile(f, "w", compression=compression) as zf:
        zf.writestr("_metadata/format-version.txt", CURRENT_VERSION)
        zf.writestr("_metadata/expression-set-type.txt", _ExpressionSetTypeIdentifier.from_expression_set(expression_set))
        zf.writestr("/latencies.txt", "\n".join(str(x) for x in expression_set.latencies))
        zf.writestr("/functions.txt", "\n".join(str(x) for x in expression_set.functions))
        zf.writestr("/blocks.txt",    "\n".join(str(x) for x in expression_set._block_names))

        for block_name in expression_set._block_names:
            zf.writestr(f"/{block_name}/channels.txt",  "\n".join(str(x) for x in expression_set._channels[block_name]))
            zf.writestr(f"/{block_name}/coo-coords.bytes", expression_set._data[block_name].data.coords.tobytes(order="C"))
            zf.writestr(f"/{block_name}/coo-data.bytes", expression_set._data[block_name].data.data.tobytes(order="C"))
            # The shape can be inferred, but we save it as an extra validation
            zf.writestr(f"/{block_name}/coo-shape.txt", "\n".join(str(x) for x in expression_set._data[block_name].data.shape))


def _load_data(from_path_or_file: PathType | FileType) -> tuple[version.Version, dict[str, Any]]:
    """
    Load an ExpressionSet from an open file, or the file at the specified path.

    If an open file is supplied, it should be opened in "rb" mode.
    """

    if isinstance(from_path_or_file, str):
        from_path_or_file = Path(from_path_or_file)

    # Check file version
    v: version.Version = file_version(from_path_or_file)
    if v < version.parse(CURRENT_VERSION):
        warn("This file uses an old format. Please consider re-saving the data to avoid future incompatibility.")
    # Loading old versions
    # For each, delegate to the appropriate _load_data_x_y() function, then
    # ensure the keys are set correctly.
    if v <= version.parse("0.1"):
        from kymata.io.nkg_compatibility import _load_data_0_1
        dict_0_1 = _load_data_0_1(from_path_or_file)

        # v0.1 data was stored as p values
        data = dict()
        for block, sparse_data_dict in dict_0_1["data"].items():
            sparse_data_dict["data"] = p_to_logp(sparse_data_dict["data"])

            sparse_data = COO(coords=sparse_data_dict["coords"],
                              data=sparse_data_dict["data"],
                              shape=sparse_data_dict["shape"],
                              prune=True, fill_value=0.0)
            assert sparse_data.shape == (
                len(dict_0_1["hexels"]),
                len(dict_0_1["latencies"]),
                len(dict_0_1["functions"]),
            )
            # In case there was only 1 function and we have a 2-d data matrix
            if len(sparse_data.shape) == 2:
                sparse_data = expand_dims(sparse_data)
            data[block] = sparse_data

        return_dict = {
            _Keys.channels:  {
                                 block_name: dict_0_1["hexels"]
                                 for block_name in _ExpressionSetTypeIdentifier.hexel.block_names()
                             },
            _Keys.functions: dict_0_1["functions"],
            _Keys.latencies: dict_0_1["latencies"],
            _Keys.data:      data,
            # Keys not present in v0.1
            _Keys.expressionset_type: _ExpressionSetTypeIdentifier.hexel,
        }
    elif v <= version.parse("0.2"):
        from kymata.io.nkg_compatibility import _load_data_0_2
        dict_0_2 = _load_data_0_2(from_path_or_file)

        # v0.2 data was stored as p-values
        data = dict()
        for block, sparse_data_dict in dict_0_2["data"].items():
            sparse_data_dict["data"] = p_to_logp(sparse_data_dict["data"])

            sparse_data = COO(coords=sparse_data_dict["coords"],
                              data=sparse_data_dict["data"],
                              shape=sparse_data_dict["shape"],
                              prune=True, fill_value=0.0)
            assert sparse_data.shape == (
                len(dict_0_2["channels"]),
                len(dict_0_2["latencies"]),
                len(dict_0_2["functions"]),
            )
            # In case there was only 1 function and we have a 2-d data matrix
            if len(sparse_data.shape) == 2:
                sparse_data = expand_dims(sparse_data)
            data[block] = sparse_data
        expressionset_type = dict_0_2["expressionset-type"]
        return_dict = {
            _Keys.expressionset_type: expressionset_type,
            _Keys.channels:           {
                                          block_name: dict_0_2["channels"]
                                          for block_name in _ExpressionSetTypeIdentifier(expressionset_type).block_names()
                                      },
            _Keys.functions:          dict_0_2["functions"],
            _Keys.latencies:          dict_0_2["latencies"],
            _Keys.data:               data,
        }
    elif v <= version.parse("0.3"):
        from kymata.io.nkg_compatibility import _load_data_0_3
        dict_0_3 = _load_data_0_3(from_path_or_file)

        # v0.2 data was stored as p-values
        data = dict()
        for block, sparse_data_dict in dict_0_3["data"].items():
            sparse_data_dict["data"] = p_to_logp(sparse_data_dict["data"])

            sparse_data = COO(coords=sparse_data_dict["coords"],
                              data=sparse_data_dict["data"],
                              shape=sparse_data_dict["shape"],
                              prune=True, fill_value=0.0)
            assert sparse_data.shape == (
                len(dict_0_3["channels"]),
                len(dict_0_3["latencies"]),
                len(dict_0_3["functions"]),
            )
            # In case there was only 1 function and we have a 2-d data matrix
            if len(sparse_data.shape) == 2:
                sparse_data = expand_dims(sparse_data)
            data[block] = sparse_data
        expressionset_type = dict_0_3["expressionset-type"]
        return_dict = {
            _Keys.expressionset_type: expressionset_type,
            _Keys.channels:           {
                                          block_name: dict_0_3["channels"]
                                          for block_name in _ExpressionSetTypeIdentifier(expressionset_type).block_names()
                                      },
            _Keys.functions:          dict_0_3["functions"],
            _Keys.latencies:          dict_0_3["latencies"],
            _Keys.data:               data,
        }
    else:
        return_dict = _load_data_current(from_path_or_file)

    return v, return_dict


# noinspection DuplicatedCode
def _load_data_current(from_path_or_file: PathType | FileType) -> dict[str, Any]:
    """
    Load data from current version
    """
    return_dict = dict()
    with open_or_use(from_path_or_file, mode="rb") as archive, ZipFile(archive, "r") as zf:

        with TextIOWrapper(zf.open("_metadata/expression-set-type.txt"), encoding="utf-8") as f:
            return_dict[_Keys.expressionset_type] = str(f.read()).strip()
        with TextIOWrapper(zf.open("/blocks.txt"), encoding="utf-8") as f:
            blocks = [str(line.strip()) for line in f.readlines()]
        with TextIOWrapper(zf.open("/latencies.txt"), encoding="utf-8") as f:
            return_dict[_Keys.latencies] = [LatencyDType(lat.strip()) for lat in f.readlines()]
        with TextIOWrapper(zf.open("/functions.txt"), encoding="utf-8") as f:
            return_dict[_Keys.functions] = [FunctionNameDType(fun.strip()) for fun in f.readlines()]
        return_dict[_Keys.channels] = dict()
        return_dict[_Keys.data] = dict()
        for block_name in blocks:
            with TextIOWrapper(zf.open(f"/{block_name}/channels.txt"), encoding="utf-8") as f:
                return_dict[_Keys.channels][block_name] = [c.strip() for c in f.readlines()]
            with zf.open(f"/{block_name}/coo-coords.bytes") as f:
                coords: NDArray = frombuffer(f.read(), dtype=int).reshape((3, -1))
            with zf.open(f"/{block_name}/coo-data.bytes") as f:
                data: NDArray = frombuffer(f.read(), dtype=float)
            with TextIOWrapper(zf.open(f"/{block_name}/coo-shape.txt"), encoding="utf-8") as f:
                shape: tuple[int, ...] = tuple(int(s.strip()) for s in f.readlines())
            sparse_data = COO(coords=coords, data=data, shape=shape, prune=True, fill_value=0.0)
            # In case there was only 1 function and we have a 2-d data matrix
            if len(shape) == 2:
                sparse_data = expand_dims(sparse_data)
            assert shape == (len(return_dict[_Keys.channels][block_name]),
                             len(return_dict[_Keys.latencies]),
                             len(return_dict[_Keys.functions]))
            return_dict[_Keys.data][block_name] = sparse_data
    return return_dict



import re
from pathlib import Path
from typing import NamedTuple

import yaml
from mne.io import Raw
from matplotlib import pyplot as plt


class Point2d(NamedTuple):
    x: float
    y: float


def get_meg_sensor_xy() -> dict[str, Point2d]:
    """
    Retrieve the 2D coordinates of MEG sensors.

    This function reads the sensor locations from a predefined layout file and returns a dictionary mapping
    sensor names to their corresponding 2D coordinates.

    Returns:
    --------
    dict[str, Point2d]
        A dictionary where keys are sensor names (e.g., 'MEG1234') and values are Point2d objects representing
        the x and y coordinates of the sensors.

    Notes:
    ------
    The function expects the layout file to be located at 'kymata-core-data/sensor_locations/Vectorview-all.lout'.
    """

    d = dict()
    layout_line_re = re.compile(
        r"^\d+\t"
        r"(?P<x>-?\d+\.\d+)\t"
        r"(?P<y>-?\d+\.\d+)\t"
        r"-?\d+\.\d+\t"
        r"-?\d+\.\d+\t"
        r"(?P<sensor>MEG \d+)$"
    )
    with Path(Path(__file__).parent.parent.parent, "kymata-core-data", "sensor_locations", "Vectorview-all.lout").open("r") as layout_file:
        _ = layout_file.readline()  # First line is nothing
        for line in layout_file:
            if not line: 
                continue  # Skip blank lines
            match = layout_line_re.match(line)
            sensor = match.group("sensor")
            sensor = sensor.replace(" ", "")
            d[sensor] = Point2d(float(match.group("x")), float(match.group("y")))
    return d


def get_eeg_sensor_xy() -> dict[str, Point2d]:
    """
    Retrieve the 2D coordinates of EEG sensors.

    This function reads the sensor locations and mappings from predefined layout and mapping files,
    then returns a dictionary mapping our sensor names to their corresponding 2D coordinates.

    Returns:
    --------
    dict[str, Point2d]
        A dictionary where keys are our sensor names and values are Point2d objects representing
        the x and y coordinates of the sensors.

    Notes:
    ------
    The function expects the layout file to be located at 'kymata-core-data/sensor_locations/EEG1005.lay'
    and the mapping file to be located at 'kymata-core-data/sensor_locations/EEG-layout-channel-mappings.yaml'.
    """
    with Path(Path(__file__).parent.parent.parent, "kymata-core-data", "sensor_locations",
              "EEG-layout-channel-mappings.yaml").open("r") as eeg_name_mapping_file:
        mapping = yaml.safe_load(eeg_name_mapping_file)
    mapping = {k.upper(): v.upper() for k, v in mapping.items()}
    d = dict()
    with Path(Path(__file__).parent.parent.parent, "kymata-core-data", "sensor_locations",
              "EEG1005.lay").open("r") as layout_file:
        for line in layout_file:
            parts = line.strip().split("\t")
            x = float(parts[1])
            y = float(parts[2])
            name = parts[-1].upper()
            d[name] = Point2d(x, y)
    our_sensor_d = {
        our_name: d[their_name]
        for our_name, their_name in mapping.items()
    }
    return our_sensor_d


def get_meg_sensors() -> set[str]:
    """Set of MEG sensor names."""
    return set(get_meg_sensor_xy().keys())


def get_eeg_sensors() -> set[str]:
    """Set of EEG sensor names."""
    return set(get_eeg_sensor_xy().keys())


def plot_eeg_sensor_positions(raw_fif: Raw):
    """
    Plot the positions of EEG sensors in 2D and 3D views.

    This function generates a figure with two subplots: a 2D view and a 3D view of the EEG sensor positions.

    Parameters:
    -----------
    raw_fif : Raw
        The raw FIF file containing EEG data and sensor locations.

    Notes:
    ------
    The 3D plot is initialized with an azimuth angle of 70 degrees and an elevation angle of 15 degrees for better visualization.
    """

    fig = plt.figure()
    ax2d = fig.add_subplot(121)
    ax3d = fig.add_subplot(122, projection='3d')
    raw_fif.plot_sensors(ch_type='eeg', axes=ax2d)
    raw_fif.plot_sensors(ch_type='eeg', axes=ax3d, kind='3d')
    ax3d.view_init(azim=70, elev=15)



from os.path import isfile
from pathlib import Path
from typing import Optional
from logging import getLogger
from warnings import warn

import numpy as np
from numpy.typing import NDArray
import mne

from kymata.io.file import PathType
from kymata.preproc.premorph import pick_channels_inverse_operator, premorph_inverse_operator

_logger = getLogger(__name__)


def load_single_emeg(emeg_path: Path,
                     need_names=False,
                     inverse_operator_path: Optional[Path] = None,
                     snr=4,
                     morph_path: Optional[Path] = None,
                     old_morph=False,
                     premorphed_inverse_operator_path: Optional[Path] = None,
                     ch_names_path: Optional[Path] = None,
                     ) -> tuple[NDArray, list[str]]:
    """
    When using the inverse operator, returns left and right hemispheres concatenated.

    old_morph: forces loading of mne morph map
    """
    emeg_path_npy = emeg_path.with_suffix(".npy")
    emeg_path_fif = emeg_path.with_suffix(".fif")

    if inverse_operator_path is None:
        ch_names_path = Path(emeg_path.parent, "ch_names.npy")

    if isfile(emeg_path_npy) and (not need_names) and (inverse_operator_path is None) and (morph_path is None):
        # Load npy-format sensor data
        channel_names: list[str] = np.load(ch_names_path)
        emeg = np.load(emeg_path_npy)

        return emeg, channel_names

    _logger.info(f"Reading EMEG evokeds from {emeg_path_fif}")
    evoked = mne.read_evokeds(emeg_path_fif, verbose=False)
    assert len(evoked) == 1
    evoked = evoked[0]

    if inverse_operator_path is None:
        # Want sensor data
        emeg = evoked.get_data()  # numpy array shape (sensor_num, N) = (370, 403_001)
        return emeg, evoked.ch_names

    if old_morph:
        # Load and apply fif-format morph data
        _logger.info(f"Reading source morph from {morph_path}")
        morph_map = mne.read_source_morph(morph_path) if morph_path is not None else None

        lh_emeg, rh_emeg, morph_hexel_names = inverse_operate(evoked, inverse_operator_path, snr, morph_map=morph_map)
        # Stack into a single matrix, to be split after gridsearch
        emeg = np.concatenate((lh_emeg, rh_emeg), axis=0)

        return emeg, morph_hexel_names

    if premorphed_inverse_operator_path is not None:
        common_channels_path = Path(premorphed_inverse_operator_path.parent, Path(premorphed_inverse_operator_path.name).stem + "_list_of_common_channels")
        if not Path(premorphed_inverse_operator_path).exists():
            # Compute premorphed operator path
            inverse_operator, premorphed_inverse_operator, morph_hexel_names = premorph_inverse_operator(
                morph_path, evoked, inverse_operator_path, snr ** -2, 'MNE', pick_ori='normal')

            # Common channels to restrict to
            common_channels = pick_channels_inverse_operator(evoked.ch_names, inverse_operator)

            np.save(premorphed_inverse_operator_path, premorphed_inverse_operator)
            np.save(ch_names_path, morph_hexel_names)
            np.save(common_channels_path, common_channels)

        else:
            # Load precomputed
            premorphed_inverse_operator = np.load(premorphed_inverse_operator_path)
            morph_hexel_names = np.load(ch_names_path, allow_pickle=True)
            common_channels = np.load(common_channels_path.with_suffix('.npy'))

        emeg = np.matmul(premorphed_inverse_operator,
                         # Restrict to common channels
                         evoked.data[common_channels])

        del evoked, premorphed_inverse_operator

    else:
        raise ValueError("Please supply premorphed_inverse_operator_path or old_morph.")

    return emeg, morph_hexel_names


def inverse_operate(evoked, inverse_operator, snr=4, morph_map: Optional[mne.SourceMorph] = None):
    lambda2 = 1.0 / snr ** 2
    _logger.info(f"Reading inverse operator from {inverse_operator}")
    inverse_operator = mne.minimum_norm.read_inverse_operator(inverse_operator, verbose=False)
    _logger.info("Applying inverse operator")
    stc: mne.VectorSourceEstimate = mne.minimum_norm.apply_inverse(evoked, inverse_operator, lambda2, 'MNE', pick_ori='normal', verbose=False)
    _logger.info("Inverse operator applied")

    if morph_map is not None:
        stc = apply_morph_map(morph_map, stc)

    return stc.lh_data, stc.rh_data, stc.vertices


def apply_morph_map(morph_map: mne.SourceMorph, stc: mne.VectorSourceEstimate):
    _logger.info("Applying morph map")
    # morph_map.apply() is very slow in some cases, for unknown reasons.
    # So we instead use a copied, patched version of the same code to make it faster,
    # at the cost of somewhat increased memory usage.
    # stc = morph_map.apply(stc)
    stc = __morph_apply(morph_map, stc)
    return stc


def __morph_apply(morph: mne.SourceMorph, stc_from, output="stc", mri_resolution=False, mri_space=None, verbose=None):
    """A copy of mne.SourceMorph.apply, for optimisation."""
    import copy
    from mne.morph import _morphed_stc_as_volume
    # _validate_type(output, str, "output")
    # _validate_type(stc_from, _BaseSourceEstimate, "stc_from", "source estimate")
    # if isinstance(stc_from, _BaseSurfaceSourceEstimate):
    #     allowed_kinds = ("stc",)
    #     extra = "when stc is a surface source estimate"
    # else:
    #     allowed_kinds = ("stc", "nifti1", "nifti2")
    #     extra = ""
    # _check_option("output", output, allowed_kinds, extra)
    stc = copy.deepcopy(stc_from)

    mri_space = mri_resolution if mri_space is None else mri_space
    if stc.subject is None:
        stc.subject = morph.subject_from
    if morph.subject_from is None:
        morph.subject_from = stc.subject
    if stc.subject != morph.subject_from:
        raise ValueError(
            "stc_from.subject and "
            "morph.subject_from must match. (%s != %s)"
            % (stc.subject, morph.subject_from)
        )
    out = __mne_apply_morph_data(morph, stc)
    if output != "stc":  # convert to volume
        out = _morphed_stc_as_volume(
            morph,
            out,
            mri_resolution=mri_resolution,
            mri_space=mri_space,
            output=output,
        )
    return out


def __mne_apply_morph_data(morph, stc_from):
    """A copy of mne.morph._apply_morph_data, for optimisation."""
    from mne.morph import _BaseSurfaceSourceEstimate, _BaseVolSourceEstimate, _check_vertices_match, \
        _VOL_MAT_CHECK_RATIO
    if stc_from.subject is not None and stc_from.subject != morph.subject_from:
        raise ValueError(
            "stc.subject (%s) != morph.subject_from (%s)"
            % (stc_from.subject, morph.subject_from)
        )
    # _check_option("morph.kind", morph.kind, ("surface", "volume", "mixed"))
    # if morph.kind == "surface":
    #     _validate_type(
    #         stc_from,
    #         _BaseSurfaceSourceEstimate,
    #         "stc_from",
    #         "volume source estimate when using a surface morph",
    #     )
    # elif morph.kind == "volume":
    #     _validate_type(
    #         stc_from,
    #         _BaseVolSourceEstimate,
    #         "stc_from",
    #         "surface source estimate when using a volume morph",
    #     )
    # else:
    #     assert morph.kind == "mixed"  # can handle any
    #     _validate_type(
    #         stc_from,
    #         _BaseSourceEstimate,
    #         "stc_from",
    #         "source estimate when using a mixed source morph",
    #     )

    # figure out what to actually morph
    do_vol = not isinstance(stc_from, _BaseSurfaceSourceEstimate)
    do_surf = not isinstance(stc_from, _BaseVolSourceEstimate)

    vol_src_offset = 2 if do_surf else 0
    from_surf_stop = sum(len(v) for v in stc_from.vertices[:vol_src_offset])
    to_surf_stop = sum(len(v) for v in morph.vertices_to[:vol_src_offset])
    from_vol_stop = stc_from.data.shape[0]
    vertices_to = morph.vertices_to
    if morph.kind == "mixed":
        vertices_to = vertices_to[0 if do_surf else 2: None if do_vol else 2]
    to_vol_stop = sum(len(v) for v in vertices_to)

    mesg = "Ori × Time" if stc_from.data.ndim == 3 else "Time"
    data_from = np.reshape(stc_from.data, (stc_from.data.shape[0], -1))
    n_times = data_from.shape[1]  # oris treated as times
    data = np.empty((to_vol_stop, n_times), stc_from.data.dtype)
    to_used = np.zeros(data.shape[0], bool)
    from_used = np.zeros(data_from.shape[0], bool)
    if do_vol:
        stc_from_vertices = stc_from.vertices[vol_src_offset:]
        vertices_from = morph._vol_vertices_from
        for ii, (v1, v2) in enumerate(zip(vertices_from, stc_from_vertices)):
            _check_vertices_match(v1, v2, "volume[%d]" % (ii,))
        from_sl = slice(from_surf_stop, from_vol_stop)
        assert not from_used[from_sl].any()
        from_used[from_sl] = True
        to_sl = slice(to_surf_stop, to_vol_stop)
        assert not to_used[to_sl].any()
        to_used[to_sl] = True
        # Loop over time points to save memory
        if morph.vol_morph_mat is None and n_times >= _VOL_MAT_CHECK_RATIO * (
                to_vol_stop - to_surf_stop
        ):
            warn(
                "Computing a sparse volume morph matrix will save time over "
                "directly morphing, calling morph.compute_vol_morph_mat(). "
                "Consider (re-)saving your instance to disk to avoid "
                "subsequent recomputation."
            )
            morph.compute_vol_morph_mat()
        if morph.vol_morph_mat is None:
            _logger.debug("Using individual volume morph")
            data[to_sl, :] = morph._morph_vols(data_from[from_sl], mesg)
        else:
            _logger.debug("Using sparse volume morph matrix")
            data[to_sl, :] = morph.vol_morph_mat @ data_from[from_sl]
    if do_surf:
        for hemi, v1, v2 in zip(
                ("left", "right"), morph.src_data["vertices_from"], stc_from.vertices[:2]
        ):
            _check_vertices_match(v1, v2, "%s hemisphere" % (hemi,))
        from_sl = slice(0, from_surf_stop)
        assert not from_used[from_sl].any()
        from_used[from_sl] = True
        to_sl = slice(0, to_surf_stop)
        assert not to_used[to_sl].any()
        to_used[to_sl] = True
        #
        #
        # Our modification
        # data[to_sl] = morph.morph_mat * data_from[from_sl]
        data[to_sl] = morph.morph_mat.todense() @ data_from[from_sl]
        #
        #
        #
    assert to_used.all()
    assert from_used.all()
    data.shape = (data.shape[0],) + stc_from.data.shape[1:]
    klass = stc_from.__class__
    stc_to = klass(data, vertices_to, stc_from.tmin, stc_from.tstep, morph.subject_to)
    return stc_to


def load_emeg_pack(emeg_filenames,
                   emeg_dir: PathType,
                   inverse_operator_dir: Optional[PathType],
                   ch_names_path: Path,
                   morph_dir: Optional[PathType] = None,
                   need_names=False,
                   ave_mode=None,
                   inverse_operator_suffix=None,
                   snr=4,
                   old_morph=False,
                   invsol_npy_dir=None,
                   ):
    """

    Args:
        emeg_filenames:
            Either: a list of reps for a single participant,
            Or: a list of the rep-average (-ave) for all participants
        emeg_dir:
        inverse_operator_dir:
        ch_names_path:
        morph_dir:
        need_names:
        ave_mode: "ave" or "concatenate"
            "ave" = average over all repetitions
            "concatenate" = treat all repetitions as if it's a single continuous stimulus
        inverse_operator_suffix:
        snr:
        old_morph:
        invsol_npy_dir:

    Returns:
        emeg: NDArray
        emeg_names: list[str]
        n_reps: int
            The number of repetitions present in the emeg array
    """

    emeg_paths = [
        Path(emeg_dir, emeg_fn)
        for emeg_fn in emeg_filenames
    ]
    n_reps = len(emeg_paths)
    morph_paths = [
        Path(morph_dir, f"{_strip_ave(emeg_fn)}_fsaverage_morph.h5") if morph_dir is not None else None
        for emeg_fn in emeg_filenames
    ]
    invsol_paths = [
        Path(invsol_npy_dir, f"{_strip_ave(emeg_fn)}{_strip_file_ext(inverse_operator_suffix)}_{_strip_ave(emeg_fn)}_fsaverage_morph.npy")
        for emeg_fn in emeg_filenames
    ]
    if inverse_operator_dir is not None:
        inverse_operator_paths = [
            Path(inverse_operator_dir, f"{_strip_ave(emeg_fn)}{inverse_operator_suffix}")
            for emeg_fn in emeg_filenames
        ]
    else:
        inverse_operator_paths = [None] * len(emeg_filenames)

    # Load first one
    try:
        emeg, emeg_names = load_single_emeg(emeg_paths[0], need_names, inverse_operator_paths[0], snr, morph_paths[0],
                                            old_morph=old_morph, premorphed_inverse_operator_path=invsol_paths[0],
                                            ch_names_path=ch_names_path)
    except Exception as ex:
        _logger.error(f"Error loading EMEG data from {str(emeg_paths[0])}")
        _logger.error(f"\tinverse operator {str(inverse_operator_paths[0])} or {str(invsol_paths[0])}")
        _logger.error(f"\tmorph {str(morph_paths[0])}")
        raise ex
    emeg = np.expand_dims(emeg, 1)

    # Load remaining ones in using the appropriate ave_mode
    if ave_mode == 'concatenate':
        # Concatenating all reps (or participant_averages - although this would be a non-standard use) into a single long stimulus
        for i in range(1, len(emeg_paths)):
            new_emeg, _ch_names = load_single_emeg(emeg_paths[i], need_names, inverse_operator_paths[i], snr,
                                                   morph_paths[i], old_morph=old_morph, premorphed_inverse_operator_path=invsol_paths[i],
                                                   ch_names_path=ch_names_path)
            emeg = np.concatenate((emeg, np.expand_dims(new_emeg, 1)), axis=1)

    elif ave_mode == 'ave':
        # Averaging together all participants

        for i in range(1, len(emeg_paths)):
            new_emeg, _ch_names = load_single_emeg(emeg_paths[i], need_names, inverse_operator_paths[i], snr,
                                                   morph_paths[i], old_morph=old_morph, premorphed_inverse_operator_path=invsol_paths[i],
                                                   ch_names_path=ch_names_path)
            emeg += np.expand_dims(new_emeg, 1)

        n_reps = 1 # n_reps is now 1 as all averaged

    elif len(emeg_paths) > 1:
        raise NotImplementedError(f'ave_mode "{ave_mode}" not known')

    return emeg, emeg_names, n_reps


def _strip_ave(name: str) -> str:
    if name.endswith("-ave"):
        return name[:-4]
    else:
        return name

def _strip_file_ext(name: str) -> str:
    """Returns a string, minus the final `.` and anything following it."""
    parts = name.split(".")
    return ".".join(parts[:-1])



from __future__ import annotations

from numpy import nanmin, greater
from numpy.typing import NDArray
import sparse
from xarray import DataArray


def expand_dims(x: sparse.COO, axis=-1) -> sparse.COO:
    """
    Expand the dims of a sparse.COO matrix by inserting a size-1 dim at the specified axis.
    axis=-1 means to add the dim at the end.
    Raises a ValueError if x isn't a 2-dimensional matrix
    """
    shape = list(x.shape)
    if len(shape) > 3:
        raise ValueError("Too many dimensions")
    if len(shape) == 3:
        if shape[2] != 1:
            raise ValueError("Too many dimensions")
        # Already expanded
        return x
    if axis == -1:
        shape.append(1)
    else:
        shape[axis:axis] = [1]
    x = x.reshape(tuple(shape))
    return x


def sparsify_log_pmatrix(log_pmatrix: NDArray) -> sparse.COO:
    """
    Converts a channel-x-latency data matrix containing log p-values into a sparse matrix
    only storing the minimum (over latencies) log p-value for each channel.
    """
    fill_value = 0.0
    channel_min = nanmin(log_pmatrix, axis=1)
    channel_min = expand_dims(channel_min, axis=2)
    # Minimum value over all latencies for this channel
    log_pmatrix[greater(log_pmatrix, channel_min)] = fill_value
    return sparse.COO(log_pmatrix, fill_value=fill_value)


def densify_data_block(x: DataArray) -> None:
    """
    Converts data in an xarray.DataArray wrapping a sparse.COO matrix to a dense numpy array.
    Operates in-place.
    """
    x.data = x.data.todense()



from matplotlib.colors import to_rgb, to_hex
from numpy import linspace


def gradient_color_dict(functions: list[str], start_color, stop_color) -> dict[str, str]:
    """
    Generates a dictionary mapping function names to colors forming a linear gradient.

    Parameters:
    -----------
    functions : list[str]
        A list of function names to be assigned colors.
    start_color : str or tuple
        The starting color of the gradient. Can be any format supported by matplotlib (e.g., color name, hex string, RGB tuple).
    stop_color : str or tuple
        The ending color of the gradient. Can be any format supported by matplotlib.

    Returns:
    --------
    dict[str, str]
        A dictionary where each key is a function name from the input list and each value is a color in hex format.
        The colors form a linear gradient between the start_color and stop_color.

    Notes:
    ------
    The returned dictionary is compatible with color specifications for expression plots.
    """
    start_rgb = to_rgb(start_color)
    stop_rgb = to_rgb(stop_color)

    return {
        function: to_hex((
            linspace(start_rgb[0], stop_rgb[0], len(functions))[i],  # R
            linspace(start_rgb[1], stop_rgb[1], len(functions))[i],  # G
            linspace(start_rgb[2], stop_rgb[2], len(functions))[i],  # B
        ))
        for i, function in enumerate(functions)
    }


def constant_color_dict(functions: list[str], color) -> dict[str, str]:
    """
    Generates a dictionary mapping function names to a single specified color.

    Parameters:
    -----------
    functions : list[str]
        A list of function names to be assigned the same color.
    color : str or tuple
        The color to assign to all function names. Can be any format supported by matplotlib (e.g., color name, hex string, RGB tuple).

    Returns:
    --------
    dict[str, str]
        A dictionary where each key is a function name from the input list and each value is the specified color in hex format.

    Notes:
    ------
    The returned dictionary is compatible with color specifications for expression plots.
    """
    return gradient_color_dict(functions, color, color)



from dataclasses import dataclass

from numpy.typing import NDArray


@dataclass
class Function:
    name: str
    values: NDArray
    sample_rate: float  # Hertz

    def downsampled(self, rate: int):
        return Function(
            name=self.name,
            values=self.values[::rate],
            sample_rate=self.sample_rate / rate,
        )

    @property
    def time_step(self) -> float:
        return 1 / self.sample_rate



from logging import getLogger
from pathlib import Path
from typing import Optional

import numpy as np
from numpy.typing import NDArray, ArrayLike
from scipy import stats

from kymata.entities.functions import Function
from kymata.math.combinatorics import generate_derangement
from kymata.math.vector import normalize, get_stds
from kymata.entities.expression import ExpressionSet, SensorExpressionSet, HexelExpressionSet
from kymata.math.p_values import log_base, p_to_logp
from kymata.plot.plot import plot_top_five_channels_of_gridsearch

_logger = getLogger(__name__)


def do_gridsearch(
        emeg_values: NDArray,  # chan x time
        function: Function,
        channel_names: list,
        channel_space: str,
        start_latency: float,   # ms
        emeg_t_start: float,    # ms
        stimulus_shift_correction: float,  # seconds/second
        stimulus_delivery_latency: float,  # seconds
        plot_location: Optional[Path] = None,
        emeg_sample_rate: int = 1000,  # Hertz
        n_derangements: int = 1,
        seconds_per_split: float = 1,
        n_splits: int = 400,
        n_reps: int = 1,
        plot_top_five_channels: bool = False,
        overwrite: bool = True,
) -> ExpressionSet:
    """
    Perform a grid search over all hexels for all latencies using EMEG data and a given function.

    This function processes EMEG data to compute the correlation between sensor or source signals
    and a specified function across multiple latencies. The results include statistical significance
    testing and optional plotting.

    Args:
        emeg_values (NDArray): A 2D array of EMEG values with shape (n_channels, time).
        function (Function): The function against which the EMEG data will be correlated. It should
            have a `values` attribute representing the function's values and a `sample_rate`
            attribute indicating its sample rate.
        channel_names (list): List of channel names corresponding to the EMEG data. For 'sensor' space,
            it is a flat list of sensor names. For 'source' space, it is a list containing two lists:
            left hemisphere and right hemisphere hexel names.
        channel_space (str): The type of channel space used, either 'sensor' or 'source'.
        start_latency (float): The starting latency for the grid search in milliseconds.
        emeg_t_start (float): The starting time of the EMEG data in milliseconds.
        stimulus_shift_correction (float): Correction factor for stimulus shift in seconds per second.
        stimulus_delivery_latency (float): Correction offset for stimulus delivery in seconds.
        plot_location (Optional[Path], optional): Path to save the plot of the top five channels of the
            grid search. If None, plotting is skipped. Default is None.
        emeg_sample_rate (int, optional): The sample rate of the EMEG data in Hertz. Default is 1000 Hz.
        n_derangements (int, optional): Number of derangements (random permutations) used to create the
            null distribution. Default is 1.
        seconds_per_split (float, optional): Duration of each split in seconds. Default is 0.5 seconds.
        n_splits (int, optional): Number of splits used for analysis. Default is 800.
        n_reps (int, optional): Number of repetitions for each split. Default is 1.
        plot_top_five_channels (bool, optional): Plots the p-values and correlation values of the top
            five channels in the gridsearch. Default is False.
        overwrite (bool, optional): Whether to overwrite existing plot files. Default is True.

    Returns:
        ExpressionSet: An ExpressionSet object (either SensorExpressionSet or HexelExpressionSet)
        containing the log p-values for each channel/hexel and latency.

    Notes:
        - The function down-samples the EMEG data to match the function's sample rate.
        - The EMEG data is reshaped into segments of the specified duration (`seconds_per_split`).
        - Cross-correlations between the EMEG data and the function are computed using FFT.
        - Statistical significance is assessed using a vectorized Welch's t-test.
        - If specified, the results are plotted and saved to the given location.
    """

    # Set random seed to keep derangement orderings
    # deterministic between runs
    np.random.seed(17)

    channel_space = channel_space.lower()
    if channel_space not in {"sensor", "source"}:
        raise NotImplementedError(channel_space)

    # We'll need to downsample the EMEG to match the function's sample rate
    downsample_rate: int = int(emeg_sample_rate / function.sample_rate)

    n_samples_per_split = int(seconds_per_split * emeg_sample_rate * 2 // downsample_rate)

    # the number of samples in the function 'trial' which is half that needed for the EMEG
    n_func_samples_per_split = n_samples_per_split // 2

    func_length = n_splits * n_func_samples_per_split
    if func_length < function.values.shape[0]:
        _logger.warning(f'WARNING: not using full length of the file (only using {round(n_splits * seconds_per_split, 2)}s)')
        func = function.values[:func_length].reshape(n_splits, n_func_samples_per_split)
    else:
        func = function.values.reshape(n_splits, n_func_samples_per_split)

    # In case func contains a fully constant split, normalize will involve a divide by zero error, resulting in a nan
    # which will infect everything downstream. Rather than try and catch and fix that, we instead kick it back to the
    # invoker to say, just ensure that this can't happen.
    try:
        func = normalize(func)
    except (ZeroDivisionError, FloatingPointError) as ex:
        _logger.error("Could not normalize function.")
        _logger.error(f"It's possible that the {function.name} function contains a constant {seconds_per_split}-second "
                      "segment, which is invalid for gridsearch. Try increasing the seconds-per-split to greater than "
                      f"{seconds_per_split} seconds, and adjust `n_splits` accordingly")
        raise ex

    n_channels = emeg_values.shape[0]

    # Reshape EMEG into splits of `seconds_per_split` s
    split_initial_timesteps = [
        int(start_latency - emeg_t_start
            + round(i * emeg_sample_rate * seconds_per_split * (1 + stimulus_shift_correction))  # splits, stretched by the shift correction
            + round(stimulus_delivery_latency * emeg_sample_rate)  # correct for stimulus delivery latency delay
            )
        for i in range(n_splits)
    ]

    emeg_reshaped = np.zeros((n_channels, n_splits * n_reps, n_samples_per_split))
    for j in range(n_reps):
        for split_i in range(n_splits):
            split_start = split_initial_timesteps[split_i]
            split_stop = split_start + int(2 * emeg_sample_rate * seconds_per_split)
            emeg_reshaped[:, split_i + (j * n_splits), :] = emeg_values[:, j, split_start:split_stop:downsample_rate]

    del emeg_values

    # Derangements for null distribution
    derangements = np.zeros((n_derangements, n_splits * n_reps), dtype=int)
    for der_i in range(n_derangements):
        derangements[der_i, :] = generate_derangement(n_splits * n_reps, n_splits)
    derangements = np.vstack((np.arange(n_splits * n_reps), derangements))  # Include the identity on top

    # Fast cross-correlation using FFT
    normalize(emeg_reshaped, inplace=True)
    emeg_stds = get_stds(emeg_reshaped, n_func_samples_per_split)
    emeg_reshaped = np.fft.rfft(emeg_reshaped, n=n_samples_per_split, axis=-1)
    F_func = np.conj(np.fft.rfft(func, n=n_samples_per_split, axis=-1))
    if n_reps > 1:
        F_func = np.tile(F_func, (n_reps, 1))
    corrs = np.zeros((n_channels, n_derangements + 1, n_splits * n_reps, n_func_samples_per_split))
    for der_i, derangement in enumerate(derangements):
        deranged_emeg = emeg_reshaped[:, derangement, :]
        corrs[:, der_i] = np.fft.irfft(deranged_emeg * F_func)[:, :, :n_func_samples_per_split] / emeg_stds[:, derangement]

    del deranged_emeg, emeg_reshaped

    # In case there was a large part of the function which was constant, the corr will be undefined (nan).
    # We want p-vals here to be 1.

    # derive pvalues
    log_pvalues = _ttest(corrs)

    latencies_ms = np.linspace(start_latency, start_latency + (seconds_per_split * 1000), n_func_samples_per_split + 1)[:-1]

    if plot_top_five_channels:
        # work out autocorrelation for channel-by-channel plots
        noise = normalize(np.random.randn(func.shape[0], func.shape[1])) * 0
        noisy_func = func + noise
        normalize(noisy_func, inplace=True)

        F_noisy_func = np.fft.rfft(noisy_func, n=n_func_samples_per_split, axis=-1)
        F_func = np.conj(np.fft.rfft(func, n=n_func_samples_per_split, axis=-1))

        auto_corrs = np.fft.irfft(F_noisy_func * F_func)

        del F_func

        plot_top_five_channels_of_gridsearch(
            corrs=corrs,
            auto_corrs=auto_corrs,
            function=function,
            n_reps=n_reps,
            n_splits=n_splits,
            n_samples_per_split=n_samples_per_split,
            latencies=latencies_ms,
            save_to=plot_location,
            log_pvalues=log_pvalues,
            overwrite=overwrite,
        )

    if channel_space == "sensor":
        es = SensorExpressionSet(
            functions=function.name,
            latencies=latencies_ms / 1000,  # seconds
            sensors=channel_names,
            data=log_pvalues,
        )
    elif channel_space == "source":
        es = HexelExpressionSet(
            functions=function.name,
            latencies=latencies_ms / 1000,  # seconds
            hexels_lh=channel_names[0],
            hexels_rh=channel_names[1],
            # Unstack the data
            data_lh=log_pvalues[:len(channel_names[0]), :],
            data_rh=log_pvalues[len(channel_names[0]):, :],
        )
    else:
        raise NotImplementedError(channel_space)

    return es


def _ttest(corrs: NDArray, use_all_lats: bool = True) -> ArrayLike:
    """
    Perform a vectorized Welch's t-test on correlation matrices.

    This function calculates the two-sample Welch's t-test statistics and their corresponding
    log p-values for given correlation matrices. The test compares true correlation values against
    null (random) correlation values after applying Fisher's Z-transformation.

    Parameters:
    -----------
    corrs : NDArray
        A 4D array of correlation values with shape (n_channels, n_derangements, n_splits, t_steps).
        - n_channels: Number of channels.
        - n_derangements: Number of derangements (random permutations).
        - n_splits: Number of splits or repetitions.
        - t_steps: Number of time steps.

    use_all_lats : bool, optional
        If True, use all latencies for computing the mean and variance of null correlations.
        If False, compute mean and variance per time step. Default is True.

    Returns:
    --------
    ArrayLike
        A 2D array of log p-values with shape (n_channels, t_steps), representing the log p-values
        of the t-tests for each channel and time step.

    Notes:
    ------
    - The correlation values are first transformed using Fisher's Z-transformation.
    - The function computes the mean and variance of the transformed true and null correlations.
    - Welch's t-test is then performed in a vectorized manner to obtain the t-statistics.
    - Depending on the degrees of freedom, either the t-distribution or normal distribution is used
      to compute the log p-values.
    - The function ensures numerical stability and precision by applying log transformations where necessary.

    """
    n_channels, n_derangements, n_splits, t_steps = corrs.shape

    # Fisher Z-Transformation
    corrs_z = 0.5 * np.log((1 + corrs) / (1 - corrs))

    # Non-deranged values are on top
    true_mean = np.mean(corrs_z[:, 0, :, :], axis=1)
    true_var = np.var(corrs_z[:, 0, :, :], axis=1, ddof=1)
    true_n = n_splits

    # Recompute mean and var for null correlations
    if use_all_lats:
        rand_mean = np.mean(corrs_z[:, 1:, :, :].reshape(n_channels, -1), axis=1).reshape(n_channels, 1)
        rand_var = np.var(corrs_z[:, 1:, :, :].reshape(n_channels, -1), axis=1, ddof=1).reshape(n_channels, 1)
        rand_n = n_splits * n_derangements * t_steps
    else:
        rand_mean = np.mean(corrs_z[:, 1:, :, :].reshape(n_channels, -1, t_steps), axis=1)
        rand_var = np.var(corrs_z[:, 1:, :, :].reshape(n_channels, -1, t_steps), axis=1, ddof=1)
        rand_n = n_splits * n_derangements

    # Vectorized two-sample t-tests for all channels and time steps
    numerator = true_mean - rand_mean
    denominator = np.sqrt(true_var / true_n + rand_var / rand_n)
    df = ((true_var / true_n + rand_var / rand_n) ** 2 /
          ((true_var / true_n) ** 2 / (true_n - 1) +
           (rand_var / rand_n) ** 2 / (rand_n - 1)))

    t_stat = numerator / denominator

    if np.min(df) <= 300:
        p = stats.t.sf(np.abs(t_stat), df) * 2  # two-tailed p-value
        log_p = p_to_logp(p)
    else:
        # norm v good approx for this, (logsf for t not implemented in logspace)
        log_p = stats.norm.logsf(np.abs(t_stat)) + np.log(2)
        log_p /= np.log(log_base)  # log base correction

    return log_p



from contextlib import contextmanager
from pathlib import Path
from typing import TextIO, BinaryIO, Union

FileType = Union[TextIO, BinaryIO]
PathType = Union[str, Path]


@contextmanager
def open_or_use(path_or_file: PathType | FileType, mode: str = "r") -> FileType:
    """
    If passed a path, will open it and return the file handle, and close when done.
    if passed a file handle, will keep it open, and return it when done.

    Use like:

        with open_or_use(path) as f:
            ...

    or like:

        with open(path) as f:
            with open_or_use(f) as ff:
                ...

    (Thanks to https://stackoverflow.com/a/6783680/2883198)
    """
    if isinstance(path_or_file, str):
        path_or_file = Path(path_or_file)
    if isinstance(path_or_file, Path):
        f = file_to_close = path_or_file.open(mode)
    else:
        f = path_or_file
        file_to_close = None

    try:
        yield f
    finally:
        if file_to_close:
            file_to_close.close()



from logging import getLogger
from os import path
from pathlib import Path
import numpy as np
import nibabel as nib

import mne

_logger = getLogger(__file__)


def create_current_estimation_prerequisites(data_root_dir, config: dict):
    """
    Copy the structurals to the local Kymata folder,
    create the surfaces, the boundary element model solutions, and the source space
    """

    list_of_participants = config['participants']
    dataset_directory_name = config['dataset_directory_name']
    interim_preprocessing_directory_name = Path(data_root_dir, dataset_directory_name, "interim_preprocessing_files")
    # mri_structural_type = config['mri_structural_type']
    mri_structurals_directory = Path(data_root_dir, dataset_directory_name, config['mri_structurals_directory'])

    '''    

    # <--------------------Command Line-------------------------->

    # Set location in the Kymata Project directory
    # where the converted MRI structurals will reside, and create the folder structure
    $ freesurfer_6.0.0
    $ setenv SUBJECTS_DIR /imaging/projects/cbu/kymata/data/dataset_4-english-narratives/raw_mri_structurals/
    for all participants:
        $ mksubjdirs participant_01 # note - this appears to ignore SUBJECTS_DIR and uses the folder you are in.

    # Load the fsaverage mesh
    $ cp -r $FREESURFER_HOME/subjects/fsaverage $SUBJECTS_DIR/fsaverage
    
    # create a one-off source space of fsaverage to morph to later.
    src = mne.setup_source_space(
        "fsaverage", spacing="ico5", subjects_dir=mri_structurals_directory, verbose=True
    )
    mne.write_source_spaces(Path(interim_preprocessing_directory_name,
                                     "4_hexel_current_reconstruction",
                                     "src_files",
                                     "fsaverage_ico5-src.fif"), src)
    
    # todo - we were using the "aparc.DKTatlas40" atlas - is this the same as the Desikan-Killiany Atlas, and aparc.DKTatlas?
    # (?h.aparc.annot)? https://surfer.nmr.mgh.harvard.edu/fswiki/CorticalParcellation. And if so, do we
    # need to add it to the fsaverage folder for fsaverage? (I don't think it is used here, although we do
    # use it in Kymata web. If so, remove this section). i.e.
    #$ cp $FREESURFER_HOME/average/rh.DKTatlas40.gcs  $SUBJECTS_DIR/fsaverage/rh.DKTatlas40.gcs
    #$ cp $FREESURFER_HOME/average/lh.DKTatlas40.gcs  $SUBJECTS_DIR/fsaverage/lh.DKTatlas40.gcs
    #$ mris_ca_label -orig white -novar fsaverage rh sphere.reg $SUBJECTS_DIR/fsaverage/label/rh.DKTatlas40.gcs $SUBJECTS_DIR/fsaverage/label/rh.aparc.DKTatlas40.annot
    #$ mris_ca_label -orig white -novar fsaverage lh sphere.reg $SUBJECTS_DIR/fsaverage/label/lh.DKTatlas40.gcs $SUBJECTS_DIR/fsaverage/label/lh.aparc.DKTatlas40.annot

    # move data across from the MRIdata folder to the local
    # directory, so freesurfer can find it - also convert from dcm to .mgz
    for participant in participants
        $ mri_convert /mridata/cbu/CBU230790_MEG23008/20231102_130449/Series005_CBU_MPRAGE_32chn/1.3.12.2.1107.5.2.43.67035.202311021312263809335255.dcm $SUBJECTS_DIR/participant_01/mri/orig/001.mgz

    # creates suitable T1, meshes and labels
    for participant in participants
        $ recon-all -s participant_01 -all

    # creates suitable T1, meshes and labels... but using python?
    for participant in participants
        The source space (downsampled version of the cortical surface in Freesurfer), which will be saved in a file ending in *-src.fif, which can be read into Matlab using mne_read_source_spaces.


        mne.viz.plot_alignment()
        mne.viz.plot_bem(),

        # create labels for these individuals, for Kymata we prefer the aparc.DKTatlas40 Atlas
    for participant in participants

        cd ${path}${subjects[m]} / label /
        mkdir Destrieux_Atlas
        mkdir DK_Atlas
        mkdir DKT_Atlas

        cd
        Destrieux_Atlas
        mne_annot2labels --subject ${subjects[m]} --parc aparc.a2009s

        cd ../DK_Atlas
        mne_annot2labels --subject ${subjects[m]} --parc aparc

        cd ../DKT_Atlas  # this is the best one to use (at the moment - from the mind-boggle dataset)
        mne_annot2labels --subject ${subjects[m]} --parc aparc.DKTatlas40

    # export to .stl file format, to offer it to participant for 3d printing (if requested)
    for participant in participants
        $ mkdir $SUBJECTS_DIR/participant_01/surf/stl_export_for_3d_printing
        $ mris_convert $SUBJECTS_DIR/participant_01/surf/rh.pial $SUBJECTS_DIR/participant_01/surf/stl_export_for_3d_printing/rh.pial.stl
        $ mris_convert $SUBJECTS_DIR/participant_01/surf/lh.pial $SUBJECTS_DIR/participant_01/surf/stl_export_for_3d_printing/lh.pial.stl
        $ zip $SUBJECTS_DIR/participant_01/surf/stl_export_for_3d_printing/participant_01_stls.zip -q -9 $SUBJECTS_DIR/participant_01/surf/stl_export_for_3d_printing/lh.pial.stl $SUBJECTS_DIR/participant_01/surf/stl_export_for_3d_printing/rh.pial.stl

    #<------------------------------------------------------------->
    '''
    # visualise the labels on the pial surface
    # for participant in list_of_participants:
    #    Brain = mne.viz.get_brain_class() # get correct brain class - why is it not doing this automatically?
    #    brain = Brain(participant, hemi="lh", surf="pial", subjects_dir=mri_structurals_directory, size=(800, 600))
    #    brain.add_annotation("aparc.a2009s", borders=False)

    # Computing the 'BEM' surfaces (needed for coregistration to work)
    for participant in list_of_participants:
        #        # andy is using:
        #        https://imaging.mrc-cbu.cam.ac.uk/meg/AnalyzingData/MNE_MRI_processing
        #
        #        if mri_structural_type == 'T1':
        #            mne.bem.make_watershed_bem(  # for T1; for FLASH, use make_flash_bem instead
        #                subject=participant,
        #                subjects_dir=mri_structurals_directory,
        #                copy=True,
        #                overwrite=True,
        #                show=True,
        #            )
        #
        #            mne.bem.make_scalp_surfaces(
        #                subject=participant,
        #                subjects_dir=mri_structurals_directory,
        #                no_decimate=True,
        #                force=True,
        #                overwrite=True,
        #            )
        #
        #        elif mri_structural_type == 'Flash':
        #            # mne.bem.make_flash_bem().
        #            print("Flash not yet implemented.")

        # produce the source space (downsampled version of the cortical surface in Freesurfer), which
        # will be saved in a file ending in *-src.fif
        src = mne.setup_source_space(
            participant, spacing="ico5", add_dist=True, subjects_dir=mri_structurals_directory
        )
        mne.write_source_spaces(Path(interim_preprocessing_directory_name,
                                     "4_hexel_current_reconstruction",
                                     "src_files",
                                     participant + "_ico5-src.fif"), src)

        fig = mne.viz.plot_bem(subject=participant,
                             subjects_dir=mri_structurals_directory,
                             brain_surfaces=["white"],
                             orientation="coronal",
                             slices=[50, 100, 150, 200])
        fig.savefig(Path(interim_preprocessing_directory_name, "4_hexel_current_reconstruction", "bem_checks", participant + "_bem_sliced.png"))

    # co-register data (make sure the MEG and EEG is aligned to the head)
    # this will save a trans .fif file
    # for participant in list_of_participants:
    #    mne.gui.coregistration(subject=participant, subjects_dir=mri_structurals_directory, block=True)

    ### #Computing the actual BEM solution
    for participant in list_of_participants:
        # conductivity = (0.3)  # for single layer
        conductivity = (0.3, 0.006, 0.3)  # for three layers
        # Note that only ico=4 is required here: Https://mail.nmr.mgh.harvard.edu/pipermail//mne_analysis/2012-June/001102.html
        model = mne.make_bem_model(subject=participant, ico=4,
                                   conductivity=conductivity,
                                   subjects_dir=mri_structurals_directory)
        bem_sol = mne.make_bem_solution(model)
        output_filename = participant + '-5120-5120-5120-bem-sol.fif'
        mne.bem.write_bem_solution(Path(mri_structurals_directory, participant, 'bem',
                                        output_filename), bem_sol, overwrite=True)


def create_forward_model_and_inverse_solution(data_root_dir, config: dict):
    list_of_participants = config['participants']
    dataset_directory_name = config['dataset_directory_name']

    mri_structurals_directory = Path(data_root_dir, dataset_directory_name, config['mri_structurals_directory'])

    interim_preprocessing_directory = Path(data_root_dir, dataset_directory_name, "interim_preprocessing_files")

    hexel_current_reconstruction_dir = Path(interim_preprocessing_directory, "4_hexel_current_reconstruction")
    hexel_current_reconstruction_dir.mkdir(exist_ok=True)

    coregistration_dir = Path(hexel_current_reconstruction_dir, "coregistration_files")
    coregistration_dir.mkdir(exist_ok=True)

    src_dir = Path(hexel_current_reconstruction_dir, "src_files")
    src_dir.mkdir(exist_ok=True)

    forward_sol_dir = Path(hexel_current_reconstruction_dir, "forward_sol_files")
    forward_sol_dir.mkdir(exist_ok=True)

    inverse_operator_dir = Path(hexel_current_reconstruction_dir, "inverse-operators")
    inverse_operator_dir.mkdir(exist_ok=True)

    # Compute forward solution
    for participant in list_of_participants:

        fwd = mne.make_forward_solution(
            Path(data_root_dir,
                 dataset_directory_name,
                 'raw_emeg', participant, participant +
                 '_run1_raw.fif'),  # note this file is only used for the sensor positions.
            trans=Path(coregistration_dir, participant + '-trans.fif'),
            src=Path(src_dir, participant + '_ico5-src.fif'),
            bem=Path(mri_structurals_directory, participant, "bem", participant + '-5120-5120-5120-bem-sol.fif'),
            meg=config['meg'],
            eeg=config['eeg'],
            mindist=5.0,
            n_jobs=None,
            verbose=True
        )
        print(fwd)

        # restrict forward vertices to those that make up cortex (i.e. all vertices that are in the annot file,
        # which does not include those in the medial wall)
        labels = mne.read_labels_from_annot(subject=participant,
                                            subjects_dir=mri_structurals_directory,
                                            parc='aparc.a2009s',
                                            hemi='both')
        fwd = mne.forward.restrict_forward_to_label(fwd, labels)

        if config['meg'] and config['eeg']:
            mne.write_forward_solution(Path(forward_sol_dir, participant + '-fwd.fif'), fwd=fwd, overwrite=True)
        elif config['meg']:
            mne.write_forward_solution(Path(forward_sol_dir, participant + '-megonly-fwd.fif'), fwd=fwd)
        elif config['eeg']:
            mne.write_forward_solution(Path(forward_sol_dir, participant + '-eegonly-fwd.fif'), fwd=fwd)
        else:
            raise Exception('eeg and meg in the dataset_config file cannot be both False')

    # Compute inverse operator

    for participant in list_of_participants:

        # Read forward solution
        if config['meg'] and config['eeg']:
            fwd = mne.read_forward_solution(Path(
                forward_sol_dir,
                participant + '-fwd.fif'))
        elif config['meg']:
            fwd = mne.read_forward_solution(Path(
                forward_sol_dir,
                participant + '-megonly-fwd.fif'))
        elif config['eeg']:
            fwd = mne.read_forward_solution(Path(
                forward_sol_dir,
                participant + '-eegonly-fwd.fif'))

        # Read noise covariance matrix
        if config['duration'] is None or config['cov_method'] != 'emptyroom':
            noise_cov = mne.read_cov(str(Path(
                hexel_current_reconstruction_dir,
                'noise_covariance_files',
                participant + "-" + config['cov_method'] + '-cov.fif')))
        else:
            noise_cov = mne.read_cov(str(Path(
                hexel_current_reconstruction_dir,
                'noise_covariance_files',
                participant + "-" + config['cov_method'] + str(config['duration']) + '-cov.fif')))

        # note this file is only used for the sensor positions.
        raw = mne.io.Raw(Path(
            Path(path.abspath("")),
            interim_preprocessing_directory,
            '2_cleaned',
            participant + '_run1_cleaned_raw.fif.gz'))

        inverse_operator = mne.minimum_norm.make_inverse_operator(
            raw.info,  # note this file is only used for the sensor positions.
            fwd,
            noise_cov,
            loose=0.2,
            depth=None,
            use_cps=True
        )
        if config['meg'] and config['eeg']:
            mne.minimum_norm.write_inverse_operator(
                str(Path(
                    inverse_operator_dir,
                    participant + '_ico5-3L-loose02-cps-nodepth-' + config['cov_method'] + '-inv.fif')),
                inverse_operator, overwrite=True)
        elif config['meg']:
            if config['duration'] is None:
                mne.minimum_norm.write_inverse_operator(
                    str(Path(
                        inverse_operator_dir,
                        participant + '_ico5-3L-loose02-cps-nodepth-megonly-' + config['cov_method'] + '-inv.fif')),
                    inverse_operator)
            else:
                mne.minimum_norm.write_inverse_operator(
                    str(Path(
                        inverse_operator_dir,
                        participant + '_ico5-3L-loose02-cps-nodepth-megonly-' + config['cov_method'] + str(
                            config['duration']) + '-inv.fif')),
                    inverse_operator)
        elif config['eeg']:
            mne.minimum_norm.write_inverse_operator(
                str(Path(
                    inverse_operator_dir,
                    participant + '_ico5-3L-loose02-cps-nodepth-eegonly-' + config['cov_method'] + '-inv.fif')),

                inverse_operator)

def confirm_digitisation_locations(data_root_dir, config):

    list_of_participants = config['participants']

    dataset_directory_name = config['dataset_directory_name']
    interim_preprocessing_directory = Path(data_root_dir, dataset_directory_name, "interim_preprocessing_files")
    hexel_current_reconstruction_dir = Path(interim_preprocessing_directory, "4_hexel_current_reconstruction")
    mri_structurals_directory = Path(data_root_dir, dataset_directory_name, config['mri_structurals_directory'])
    coregistration_dir = Path(hexel_current_reconstruction_dir, "coregistration_files")
    src_dir = Path(hexel_current_reconstruction_dir, "src_files")
    fwd_dir = Path(hexel_current_reconstruction_dir, "forward_sol_files")
    coregistration_checks_dir = Path(coregistration_dir, "confirmation_of_digitisation_locations")
    helmet_intersection_check_dir = Path(coregistration_checks_dir, 'helmet-intersection-check')
    sensor_location_check_dir = Path(coregistration_checks_dir, 'sensor-location-check')
    src_check_dir = Path(coregistration_checks_dir, 'src-check')
    fwd_check_dir = Path(coregistration_checks_dir, 'fwd-check')
    medial_wall_omitted_check_dir = Path(coregistration_checks_dir, 'medial-wall-omitted-check')
    side_check_dir = Path(coregistration_checks_dir, 'side-check')

    coregistration_checks_dir.mkdir(exist_ok=True)
    helmet_intersection_check_dir.mkdir(exist_ok=True)
    sensor_location_check_dir.mkdir(exist_ok=True)
    src_check_dir.mkdir(exist_ok=True)
    fwd_check_dir.mkdir(exist_ok=True)
    medial_wall_omitted_check_dir.mkdir(exist_ok=True)
    side_check_dir.mkdir(exist_ok=True)

    for participant in list_of_participants:

        raw_fname = Path(interim_preprocessing_directory, '2_cleaned', participant + '_run1_cleaned_raw.fif.gz')
        trans_fname = Path(coregistration_dir, participant + '-trans.fif')
        src_name = Path(src_dir, participant + '_ico5-src.fif')
        fwd_name = Path(fwd_dir, participant + '-fwd.fif')

        raw = mne.io.read_raw_fif(raw_fname)
        trans = mne.read_trans(trans_fname)
        src = mne.read_source_spaces(src_name)
        fwd = mne.read_forward_solution(fwd_name)

        fwd_fixed = mne.convert_forward_solution(
            fwd, surf_ori=True, force_fixed=True, use_cps=True
        )

        # Load the T1 file and change the header information to the correct units
        t1w = nib.load(Path(mri_structurals_directory, participant, "mri", "T1.mgz"))
        t1w = nib.Nifti1Image(t1w.dataobj, t1w.affine)
        t1w.header["xyzt_units"] = np.array(10, dtype="uint8")

        fig = mne.viz.plot_alignment(
            raw.info,
            trans=trans,
            subject=participant,
            subjects_dir=mri_structurals_directory,
            surfaces=["head-dense", "white"],
            show_axes=True,
            dig=True,
            meg={"helmet": 0.1,
                 "sensors": 0.1},
            coord_frame="meg",
            mri_fiducials="estimated",
        )
        mne.viz.set_3d_view(fig, 0, 90, distance=0.6, focalpoint=(0.0, 0.0, 0.0))
        fig.plotter.screenshot(Path(helmet_intersection_check_dir, participant+'-helmet-intersection-check.png'))

        fig = mne.viz.plot_alignment(
            raw.info,
            trans=trans,
            subject=participant,
            subjects_dir=mri_structurals_directory,
            surfaces=["head-dense", "white"],
            show_axes=True,
            dig=True,
            eeg=dict(original=0, projected=1),
            meg={"sensors": 0.1},
            coord_frame="meg",
            mri_fiducials="estimated",
        )
        mne.viz.set_3d_view(fig, 45, 90, distance=0.6, focalpoint=(0.0, 0.0, 0.0))
        fig.plotter.screenshot(Path(sensor_location_check_dir, participant + '-sensor-location-check.png'))

        fig = mne.viz.plot_alignment(
            raw.info,
            trans=trans,
            subject=participant,
            src=src,
            subjects_dir=mri_structurals_directory,
            surfaces=["white"],
            eeg=False,
            dig='fiducials',
            show_axes=True,
            meg={"sensors": 0.0},
            coord_frame="meg",
        )
        mne.viz.set_3d_view(fig, 45, 90, distance=0.4, focalpoint=(0.0, 0.0, 0.0))
        fig.plotter.screenshot(Path(src_check_dir, participant + '-src-check.png'))

        fig = mne.viz.plot_alignment(
            raw.info,
            trans=trans,
            subject=participant,
            dig=False,
            eeg=False,
            fwd=fwd_fixed,
            subjects_dir=mri_structurals_directory,
            surfaces=["white"],
            show_axes=True,
            meg={"sensors": 0.0},
            coord_frame="meg",
        )
        mne.viz.set_3d_view(fig, 45, 90, distance=0.4, focalpoint=(0.0, 0.0, 0.0))
        fig.plotter.screenshot(Path(fwd_check_dir, participant + '-fwd-check.png'))

        fig = mne.viz.plot_alignment(
            raw.info,
            trans=trans,
            subject=participant, 
            dig=False,
            eeg=False,
            fwd=fwd_fixed,
            subjects_dir=mri_structurals_directory,
            surfaces=["white"],
            show_axes=True,
            meg={"sensors": 0.0},
            coord_frame="meg",
        )
        mne.viz.set_3d_view(fig, 0, 180, distance=0.4, focalpoint=(0.0, 0.0, 0.0))
        fig.plotter.screenshot(Path(medial_wall_omitted_check_dir, participant + '-medial-wall-omitted-check.png'))

        fig = mne.viz.plot_alignment(
            raw.info,
            trans=trans,
            subject=participant,
            dig=False,
            eeg=False,
            fwd=fwd_fixed,
            subjects_dir=mri_structurals_directory,
            surfaces=["white"],
            show_axes=True,
            meg={"sensors": 0.0},
            coord_frame="meg",
        )
        mne.viz.set_3d_view(fig, 0, 90, distance=0.4, focalpoint=(0.0, 0.0, 0.0))
        fig.plotter.screenshot(Path(side_check_dir, participant + '-side-check.png'))



def create_hexel_morph_maps(data_root_dir, config: dict):
    list_of_participants = config['participants']
    dataset_directory_name = config['dataset_directory_name']

    mri_structurals_directory = Path(data_root_dir, dataset_directory_name, config['mri_structurals_directory'])

    interim_preprocessing_directory = Path(data_root_dir, dataset_directory_name, "interim_preprocessing_files")

    hexel_current_reconstruction_dir = Path(interim_preprocessing_directory, "4_hexel_current_reconstruction")
    hexel_current_reconstruction_dir.mkdir(exist_ok=True)

    src_dir = Path(hexel_current_reconstruction_dir, "src_files")
    src_dir.mkdir(exist_ok=True)

    forward_sol_dir = Path(hexel_current_reconstruction_dir, "forward_sol_files")
    forward_sol_dir.mkdir(exist_ok=True)

    morph_map_dir = Path(hexel_current_reconstruction_dir, "morph_maps")
    morph_map_dir.mkdir(exist_ok=True)

    for participant in list_of_participants:

        morphmap_filename = Path(morph_map_dir, participant + "_fsaverage_morph.h5")

        # First compute morph matrices for participant
        if not path.isfile(morphmap_filename):

            # read the src space not from the original but from the version in fwd or
            # inv, incase any vertices have been removed due to proximity to the scalp
            # https://mne.tools/stable/auto_tutorials/forward/30_forward.html#sphx-glr-auto-tutorials-forward-30-forward-py
            fwd = mne.read_forward_solution(Path(
                forward_sol_dir,
                participant + '-fwd.fif'))
            src_from = fwd['src']

            src_to = mne.read_source_spaces(Path(
                src_dir,
                'fsaverage_ico5-src.fif'))

            morph = mne.compute_source_morph(
                src_from,
                subject_from=participant,
                subject_to="fsaverage",
                src_to=src_to,
                subjects_dir=mri_structurals_directory,
            )
            morph.save(morphmap_filename)
        else:
            _logger.info("Morph maps already created")



from sys import stdout

from colorama import Style


def print_with_color(message: str, fore_color, style=Style.BRIGHT):
    """Print a message in a bright style"""
    print(f"{fore_color}{style}{message}{Style.RESET_ALL}")


def input_with_color(message: str, fore_color, style=Style.BRIGHT) -> str:
    """Get input in a bright style"""
    return input(f"{fore_color}{style}{message}{Style.RESET_ALL}")


def print_progress(iteration: int, total: int,
                   prefix: str = '',
                   suffix: str = '',
                   *,
                   decimals: int = 1,
                   bar_length: int = 100,
                   update_downsample: int = 1,
                   full_char='▒',
                   empty_char='┄',
                   terminal_char='║',
                   clear_on_completion: bool = False):
    """
    Call in a loop to create terminal progress bar.
    Based on https://github.com/emcoglab/ldm-core/blob/main/utils/log.py
    @params:
        iteration           - Required : current iteration (0-indexed) (Int)
                                         (so the at the last iteration, `iteration == total - 1`)
        total               - Required : total iterations (Int)
        prefix              - Optional : prefix string (Str)
        suffix              - Optional : suffix string (Str)
        decimals            - Optional : positive number of decimals in percent complete (Int)
        bar_length          - Optional : character length of bar (Int)
        clear_on_completion - Optional : clear the bar when it reaches 100% (bool)
    """

    iteration += 1  # convert to 1-indexed to make the maths easier

    if iteration < total and iteration % update_downsample != 0:
        stdout.flush()
        return

    portion_complete = iteration / float(total)
    percents = f"{100 * portion_complete:.{decimals}f}%"
    filled_length = int(round(bar_length * portion_complete))
    bar = (full_char * filled_length) + (empty_char * (bar_length - filled_length))

    stdout.write(f'\r{prefix}{terminal_char}{bar}{terminal_char} {percents}{suffix}'),

    if iteration == total:
        stdout.write("\r" if clear_on_completion else "\n")

    stdout.flush()



from __future__ import annotations

from pathlib import Path
from typing import Optional

from numpy import nan_to_num, minimum, array
from scipy.io import loadmat as loadmat_pre_73
from mat73 import loadmat as loadmat_post_73

from kymata.entities.expression import HexelExpressionSet
from kymata.math.p_values import p_to_logp
from kymata.entities.iterables import all_equal


def load_mat(path):
    """Loads all variables in a matlab file, regardless of version."""
    try:
        with path.open("rb") as f:
            mat = loadmat_pre_73(f, appendmat=False)
    except NotImplementedError:
        mat = loadmat_post_73(path)
    return mat


def load_mat_variable(path, variable_name: str):
    """Loads a specified named variable from a matlab file"""
    return load_mat(path)[variable_name]


def load_matab_hexel_expression_files(
        function_name: str,
        lh_file: Path | str,
        rh_file: Path | str,
        flipped_lh_file: Optional[Path | str] = None,
        flipped_rh_file: Optional[Path | str] = None,
) -> HexelExpressionSet:
    """Load from a set of MATLAB files."""

    if flipped_lh_file is None or flipped_rh_file is None:
        assert flipped_lh_file is None and flipped_rh_file is None, "Please supply 2 or 4 files."
        return _load_matab_expression_files_combined_flipped(
            function_name=function_name, lh_file=lh_file, rh_file=rh_file
        )
    else:
        return _load_matab_expression_files_separate_flipped(
            function_name=function_name, lh_file=lh_file, rh_file=rh_file,
            flipped_lh_file=flipped_lh_file, flipped_rh_file=flipped_rh_file
        )


def _load_matlab_validate(lh_mats: tuple[dict, ...], rh_mats: tuple[dict, ...]) -> None:
    """
    Returns silently unless there are validation errors
    """
    all_mats = (*lh_mats, *rh_mats)
    assert len(all_mats) in {2, 4}

    # All the same function
    assert all_equal([_base_function_name(mat["functionname"]) for mat in all_mats])
    # Timing information is the same
    assert all_equal([mat["latency_step"]               for mat in all_mats])
    assert all_equal([len(mat["latencies"])             for mat in all_mats])
    assert all_equal([mat["nTimePoints"]                for mat in all_mats])
    assert all_equal([mat["outputSTC"]["tmin"]          for mat in all_mats])
    assert all_equal([mat["outputSTC"]["tstep"]         for mat in all_mats])
    assert all_equal([mat["outputSTC"]["data"].shape[0] for mat in all_mats])

    assert all_mats[0]["outputSTC"]["data"].shape[0] == all_mats[0]["nTimePoints"]
    # Spatial information is the same
    assert all_equal([mat["nVertices"]                       for mat in lh_mats])
    assert all_equal([mat["nVertices"]                       for mat in rh_mats])
    assert all_equal([len(mat["outputSTC"]["vertices"])      for mat in lh_mats])
    assert all_equal([len(mat["outputSTC"]["vertices"])      for mat in rh_mats])
    assert all_equal([mat["outputSTC"]["data"].shape[1]      for mat in all_mats])
    assert all_mats[0]["outputSTC"]["data"].shape[1] == all_mats[0]["nVertices"]


def _load_matlab_downsample_ratio(lh_mat: dict) -> int:
    # If the data has been downsampled
    if lh_mat["outputSTC"]["tstep"] != lh_mat["latency_step"]/1000:
        downsample_ratio = (lh_mat["latency_step"] / 1000) / lh_mat["outputSTC"]["tstep"]
        assert downsample_ratio == int(downsample_ratio)
        downsample_ratio = int(downsample_ratio)
    else:
        downsample_ratio = 1
    return downsample_ratio


def _prep_matlab_data(data, n_latencies, downsample_ratio):
    return (
        # Some hexels are all nans because they're on the medial wall
        # or otherwise intentionally excluded from the analysis;
        # we replace those with p=1.0 to ignore
        nan_to_num(
            nan=1.0,
            x=_downsample_data(data, downsample_ratio)
            # Trim excess
            [:n_latencies, :],
        )
    )


def _load_matab_expression_files_separate_flipped(
        function_name: str,
        lh_file: Path | str, flipped_lh_file: Path | str,
        rh_file: Path | str, flipped_rh_file: Path | str) -> HexelExpressionSet:
    """
    For loading Matlab files where the flipped and non-flipped versions are separate
    (expects 4 files).
    """
    for p in [lh_file, rh_file, flipped_lh_file, flipped_rh_file]:
        if not Path(p).exists():
            raise FileNotFoundError(p)

    lh_mat = load_mat(Path(lh_file))
    rh_mat = load_mat(Path(rh_file))
    flipped_lh_mat = load_mat(Path(flipped_lh_file))
    flipped_rh_mat = load_mat(Path(flipped_rh_file))

    # Check 4 files are compatible
    assert lh_mat["leftright"] == flipped_lh_mat["leftright"] == "lh"
    assert rh_mat["leftright"] == flipped_rh_mat["leftright"] == "rh"
    _load_matlab_validate(lh_mats=(lh_mat, flipped_lh_mat), rh_mats=(rh_mat, flipped_rh_mat))

    downsample_ratio = _load_matlab_downsample_ratio(lh_mat)

    # Combine flipped and non-flipped
    # Matlab data contains p-values
    logp_matrix_lh = p_to_logp(minimum(
        _prep_matlab_data(lh_mat["outputSTC"]["data"], n_latencies=len(lh_mat["latencies"]), downsample_ratio=downsample_ratio),
        _prep_matlab_data(flipped_lh_mat["outputSTC"]["data"], n_latencies=len(lh_mat["latencies"]), downsample_ratio=downsample_ratio)
    ))
    logp_matrix_rh = p_to_logp(minimum(
        _prep_matlab_data(rh_mat["outputSTC"]["data"], n_latencies=len(lh_mat["latencies"]), downsample_ratio=downsample_ratio),
        _prep_matlab_data(flipped_rh_mat["outputSTC"]["data"], n_latencies=len(lh_mat["latencies"]), downsample_ratio=downsample_ratio)
    ))

    return HexelExpressionSet(
        functions=function_name,
        hexels_lh=lh_mat["outputSTC"]["vertices"], hexels_rh=rh_mat["outputSTC"]["vertices"],
        latencies=lh_mat["latencies"] / 1000,  # These will be the same left and right
        data_lh=array(logp_matrix_lh).T, data_rh=array(logp_matrix_rh).T,
    )


def _load_matab_expression_files_combined_flipped(
        function_name: str,
        lh_file: Path | str, rh_file: Path | str) -> HexelExpressionSet:
    """
    For loading Matlab files where the flipped and non-flipped versions are already combined
    (expects 2 files).
    """

    for p in [lh_file, rh_file]:
        if not Path(p).exists():
            raise FileNotFoundError(p)

    lh_mat = load_mat(Path(lh_file))
    rh_mat = load_mat(Path(rh_file))

    # Check 2 files are compatible
    _load_matlab_validate(lh_mats=(lh_mat,), rh_mats=(rh_mat,))
    assert lh_mat["leftright"] == "lh"
    assert rh_mat["leftright"] == "rh"

    downsample_ratio = _load_matlab_downsample_ratio(lh_mat)

    logp_matrix_lh = p_to_logp(_prep_matlab_data(lh_mat["outputSTC"]["data"],
                                                 n_latencies=len(lh_mat["latencies"]),
                                                 downsample_ratio=downsample_ratio))
    logp_matrix_rh = p_to_logp(_prep_matlab_data(rh_mat["outputSTC"]["data"],
                                                 n_latencies=len(lh_mat["latencies"]),
                                                 downsample_ratio=downsample_ratio))

    return HexelExpressionSet(
        functions=function_name,
        hexels_lh=lh_mat["outputSTC"]["vertices"], hexels_rh=rh_mat["outputSTC"]["vertices"],
        latencies=lh_mat["latencies"] / 1000,  # These will be the same left and right
        data_lh=array(logp_matrix_lh).T, data_rh=array(logp_matrix_rh).T,
    )


def _base_function_name(function_name: str) -> str:
    """
    Removes extraneous metadata from function names.
    """
    function_name = function_name.removesuffix("-flipped")
    return function_name


def _downsample_data(data, ratio):
    """
    Subsample a numpy array in the first dimension.
    """
    if ratio == 1:
        return data
    else:
        return data[::ratio, :]



import math


def round_down(val: float, interval: int):
    """
    Rounds down to the nearest multiple of `multiple`.
    E.g. round_down(13, 10) = 10
    """
    return int(math.floor(val / interval)) * interval


def round_up(val: float, interval: int):
    """
    Rounds up to the nearest multiple of `multiple`.
    E.g. round_up(13, 10) = 20
    """
    return int(math.ceil(val / interval)) * interval



from numpy import int_, str_, float_


Hexel = int
Sensor = str
Latency = float

# Set consistent dtypes for use in arrays
HexelDType = int_
SensorDType = str_
LatencyDType = float_
FunctionNameDType = str_



from typing import Dict, List, Tuple, Any

from scipy.interpolate import splev
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.lines import Line2D

from kymata.ippm.data_tools import Node


class IPPMPlotter:
    def plot(self, graph: Dict[str, Node], func_colors: Dict[str, str]):
        x_coords, y_coords, colors, sizes, edge_colors, b_splines = self._extract_data_info_from_dict(graph, func_colors)
        self._draw(x_coords, y_coords, colors, sizes, edge_colors, b_splines, func_colors)

    def _extract_data_info_from_dict(self, graph: Dict[str, Node], func_colors: Dict[str, str]):
        number_of_nodes = len(graph.keys())
        x_coords = [_ for _ in range(number_of_nodes)]
        y_coords = [_ for _ in range(number_of_nodes)]
        colors = [_ for _ in range(number_of_nodes)]
        sizes = [_ for _ in range(number_of_nodes)]
        b_splines = [_ for _ in range(number_of_nodes)]
        edge_colors = []

        for idx, node_infos in enumerate(graph.items()):
            node_name, node_info = node_infos

            function_name = node_name[:node_name.index('-')] if '-' in node_name else node_name
            colors[idx] = func_colors[function_name]

            sizes[idx] = node_info.magnitude
            x_coord, y_coord = node_info.position
            x_coords[idx], y_coords[idx] = x_coord, y_coord

            pairs = []
            for inc_edge in node_info.inc_edges:
                start_x, start_y = graph[inc_edge].position
                pairs.append([(start_x, start_y), (x_coord, y_coord)])
                edge_colors.append(colors[idx])

            b_splines.append(self._make_b_spline_paths(pairs))

        return x_coords, y_coords, colors, sizes, edge_colors, b_splines

    def _make_b_spline_paths(self, pairs: List[Tuple[Any, Any]]) -> np.array:
        """
        Given a list of hexel positions pairs, return a list of
        b-splines. First, find the control points, and second
        create the b-splines from these control points.

        Args:
            hexel_coordinate_pairs (List[List[Tuple[float, float]]]): Each list contains the x-axis values and y-axis values
                for the start and end of a BSpline, e.g., [(0, 1), (1, 0)].

        Returns:
            List[List[np.array]]: A list of a list of np arrays. Each list contains two np.arrays. The first np.array contains
                the x-axis points and the second one contains the y-axis points. Together, they define a BSpline. Thus, it is
                a list of BSplines.
        """
        bspline_path_array = []
        for pair in pairs:
            start_X = pair[0][0]
            start_Y = pair[0][1]
            end_X = pair[1][0]
            end_Y = pair[1][1]

            if start_X + 35 > end_X and start_Y == end_Y:
                # the nodes are too close to use a bspline. Null edge.
                # add 2d np array where the first element is xs and second is ys
                xs = np.linspace(start_X, end_X, 100, endpoint=True)
                ys = np.array([start_Y] * 100)
                bspline_path_array.append([xs, ys])
            else:
                ctr_pts = self._make_b_spline_ctr_points(pair)
                bspline_path_array.append(self._make_b_spline_path(ctr_pts))

        return bspline_path_array

    def _make_b_spline_ctr_points(self, start_and_end_node_coordinates: List[Tuple[float, float]]) -> np.array:
        """
        Given the position of a start hexel and an end hexel, create
        a set of 6 control points needed for a b-spline.

        The first one and last one is the position of a start hexel
        and an end hexel themselves, and the intermediate four are
        worked out using some simple rules.

        Args:
            start_and_end_node_coordinates (List[Tuple[float, float]]): List containing the start and end coordinates for one edge.
                First tuple is start, second is end. First element in tuple is x coord, second is y coord.

        Returns:
            np.array: A list of tuples of coordinates. Each coordinate pair represents a control point.
        """

        start_X = start_and_end_node_coordinates[0][0]
        start_Y = start_and_end_node_coordinates[0][1]
        end_X = start_and_end_node_coordinates[1][0]
        end_Y = start_and_end_node_coordinates[1][1]

        if end_X < start_X:
            # reverse BSpline
            start_X, end_X = end_X, start_X
            start_Y, end_Y = end_Y, start_Y

        bspline_ctr_points = []
        bspline_ctr_points.append((start_X,start_Y))

        # first 2
        bspline_ctr_points.append((start_X + 20, start_Y))
        bspline_ctr_points.append((start_X + 25, start_Y))

        # second 2
        bspline_ctr_points.append((start_X + 30, end_Y))
        bspline_ctr_points.append((start_X + 35, end_Y))

        bspline_ctr_points.append((end_X, end_Y))

        bspline_ctr_points = np.array(bspline_ctr_points)

        return bspline_ctr_points

    def _make_b_spline_path(self, ctr_points: np.array) -> List[np.array]:
        """
        With an input of six control points, return an interpolated
        b-spline path which corresponds to a curved edge from one node to another.

        Args:
            ctr_points (np.array): 2d np.array containing the coordinates of the center points.

        Returns:
            List[np.array]: A list of np.arrays that represent one BSpline path. The first list is a list of x-axis coordinates
                the second is a list of y-axis coordinates.
        """
        x = ctr_points[:, 0]
        y = ctr_points[:, 1]

        length = len(x)
        t = np.linspace(0, 1, length - 2, endpoint=True)
        t = np.append([0, 0, 0], t)
        t = np.append(t, [1, 1, 1])

        tck = [t, [x, y], 3]
        u3 = np.linspace(0, 1, (max(length * 2, 70)), endpoint=True)
        bspline_path = splev(u3, tck)

        return bspline_path

    def _draw(
            self,
            x_coords: List[float],
            y_coords: List[float],
            colors: List[str],
            sizes: List[float],
            edge_colors: List[str],
            b_splines: np.array,
            func_colors: Dict[str, str]
    ):
        fig, ax = plt.subplots()
        for path, color in zip(b_splines, edge_colors):
            ax.plot(path[0], path[1], color=color, linewidth='3', zorder=-1)

        ax.scatter(x=x_coords, y=y_coords, c=colors, s=sizes, zorder=1)
        legend = []
        for f in func_colors.keys():
            legend.append(Line2D([0], [0], marker='o', color='w', label=f, markerfacecolor=func_colors[f], markersize=15))

        plt.legend(handles=legend, loc='upper left')

        ax.set_ylim(min(y_coords) - 0.1, max(y_coords) + 0.1)
        ax.set_yticklabels([])
        ax.yaxis.set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.set_xlabel('Latency (ms)')

        fig.set_figheight(5)
        fig.set_figwidth(10)


        plt.show()


"""
Classes and functions for storing expression information.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Sequence, Union, get_args, Tuple
from warnings import warn

from numpy import array, array_equal, ndarray
from numpy.typing import NDArray
from pandas import DataFrame
from sparse import SparseArray, COO
from xarray import DataArray, concat

from kymata.entities.datatypes import HexelDType, SensorDType, LatencyDType, FunctionNameDType, Hexel, Sensor, \
    Latency
from kymata.entities.iterables import all_equal
from kymata.entities.sparse_data import expand_dims, densify_data_block, sparsify_log_pmatrix

_InputDataArray = Union[ndarray, SparseArray]  # Type alias for data which can be accepted

# Data dimension labels
DIM_HEXEL = "hexel"
DIM_SENSOR = "sensor"
DIM_LATENCY = "latency"
DIM_FUNCTION = "function"

# Block (e.g. hemisphere)
BLOCK_LEFT  = "left"
BLOCK_RIGHT = "right"
BLOCK_SCALP = "scalp"


class ExpressionSet(ABC):
    """
    Brain data associated with expression of a single function.
    Data is log10 p-values
    """

    def __init__(self,
                 functions: str | Sequence[str],
                 # Metadata
                 latencies: Sequence[Latency],
                 # In general, we will combine flipped and non-flipped versions
                 # data_blocks contains a dict mapping block names to data arrays
                 # e.g., in the case there are three functions:
                 #             "left" → [array(), array(), array()],
                 #        and "right" → [array(), array(), array()]
                 #   or:      "scalp" → [array(), array(), array()]
                 # All should be the same size.
                 data_blocks: dict[str, _InputDataArray | Sequence[_InputDataArray]],  # log p-values
                 # Supply channels already as an array, i.e.
                 channel_coord_name: str,
                 channel_coord_dtype,
                 channel_coord_values: dict[str, Sequence],
                 ):
        """
        data_blocks' values should store log10 p-values
        """

        # Canonical order of dimensions
        self._dims = (channel_coord_name, DIM_LATENCY, DIM_FUNCTION)

        self._block_names: list[str] = list(data_blocks.keys())
        self.channel_coord_name = channel_coord_name

        # Validate arguments
        assert set(self._block_names) == set(channel_coord_values.keys()), "Ensure data block names match channel block names"
        _length_mismatch_message = ("Argument length mismatch, please supply one function name and accompanying data, "
                                    "or equal-length sequences of the same.")
        if isinstance(functions, str):
            # If only one function
            for data in data_blocks.values():
                # Data should not be a sequence
                assert isinstance(data, get_args(_InputDataArray)), _length_mismatch_message
            # Wrap into sequences
            functions = [functions]
            for bn in self._block_names:
                data_blocks[bn] = [data_blocks[bn]]

        self._validate_functions_no_duplicates(functions)

        for data in data_blocks.values():
            assert len(functions) == len(data), _length_mismatch_message
        assert all_equal([arr.shape[1] for arrs in data_blocks.values() for arr in arrs]), "Not all input data blocks have the same"

        # Channels can vary between blocks (e.g. different numbers of vertices for each hemisphere).
        # But latencies and functions do not
        channels: dict[str, NDArray] = {
            bn: array(channel_coord_values[bn], dtype=channel_coord_dtype)
            for bn in self._block_names
        }
        latencies = array(latencies, dtype=LatencyDType)
        functions = array(functions, dtype=FunctionNameDType)

        # Input value `data_blocks` has type something like dict[str, list[array]].
        #  i.e. a dict mapping block names to a list of 2d data volumes, one for each function
        # We need to eventually get it into `self._data`, which has type dict[str, DataArray]
        # i.e. a dict mapping block names to a DataArray containing data for all functions
        self._data: dict[str, DataArray] = dict()
        nan_warning_sent = False
        for block_name, data_for_functions in data_blocks.items():
            for function_name, data in zip(functions, data_for_functions):
                assert len(channels[block_name]) == data.shape[0], f"{channel_coord_name} mismatch for {function_name}: {len(channels)} {channel_coord_name} versus data shape {data.shape} ({block_name})"
                assert len(latencies) == data.shape[1], f"Latencies mismatch for {function_name}: {len(latencies)} latencies versus data shape {data.shape}"
            data_array: DataArray = concat(
                (
                    DataArray(self._init_prep_data(d),
                              coords={
                                  channel_coord_name: channels[block_name],
                                  DIM_LATENCY: latencies,
                                  DIM_FUNCTION: [function]
                              })
                    for function, d in zip(functions, data_for_functions)
                ),
                dim=DIM_FUNCTION,
                data_vars="all",  # Required by concat of DataArrays
                )

            # Sometimes the data can contain nans, for example if the MEG hexel currents were set to nan on the medial
            # wall. We can ignore these nans by setting the values to p=1, but because it's not typically expected we
            # warn the user about it.
            if data_array.isnull().any():
                data_array = data_array.fillna(value=0)  # logp = 0 => p = 1
                if not nan_warning_sent:  # Only want to send the warning once, even if there are multiple data blocks with nans.
                    warn("Supplied data contained nans. These will be replaced by p = 1 values.")
                    nan_warning_sent = True

            assert data_array.dims == self._dims
            assert set(data_array.coords.keys()) == set(self._dims)
            assert array_equal(data_array.coords[DIM_FUNCTION].values, functions)

            self._data[block_name] = data_array

    def _validate_functions_no_duplicates(self, functions):
        if not len(functions) == len(set(functions)):
            checked = []
            for f in functions:
                if f in checked:
                    break
                checked.append(f)
            raise ValueError(f"Duplicated functions in input, e.g. {f}")

    @classmethod
    def _init_prep_data(cls, data: _InputDataArray) -> COO:
        if isinstance(data, ndarray):
            data = sparsify_log_pmatrix(data)
        elif not isinstance(data, SparseArray):
            raise NotImplementedError()
        data = expand_dims(data, 2)
        return data

    @property
    # block → channels
    def _channels(self) -> dict[str, NDArray]:
        return {
            bn: data.coords[self.channel_coord_name].values
            for bn, data in self._data.items()
        }

    @property
    def functions(self) -> list[FunctionNameDType]:
        """Function names."""
        functions = {
            bn: data.coords[DIM_FUNCTION].values.tolist()
            for bn, data in self._data.items()
        }
        # Validate that functions are the same for all data blocks
        assert all_equal(list(functions.values()))
        # Then just return the first one
        return functions[self._block_names[0]]

    @property
    def latencies(self) -> NDArray[LatencyDType]:
        """Latencies, in seconds."""
        latencies = {
            bn: data.coords[DIM_LATENCY].values
            for bn, data in self._data.items()
        }
        # Validate that latencies are the same for all data blocks
        assert all_equal(list(latencies.values()))
        # Then just return the first one
        return latencies[self._block_names[0]]

    @abstractmethod
    def __getitem__(self, functions: str | Sequence[str]) -> ExpressionSet:
        pass

    @abstractmethod
    def __copy__(self) -> ExpressionSet:
        pass

    @abstractmethod
    def __add__(self, other) -> ExpressionSet:
        pass

    @abstractmethod
    def __eq__(self, other: ExpressionSet) -> bool:
        # Override this method and provide additional checks after calling super().__eq__(other)
        if type(self) is not type(other):
            return False
        if not self.functions == other.functions:
            return False
        if not array_equal(self.latencies, other.latencies):
            return False
        for bn in self._block_names:
            if not array_equal(self._channels[bn], other._channels[bn]):
                return False
        return True

    def _best_functions_for_block(self, block_name: str) -> DataFrame:
        """
        Return a DataFrame containing:
        for each channel in the block, the best function and latency for that channel, and the associated log p-value

        Note that channels for which the best p-value is 1 will be omitted.
        """
        # Want, for each channel in the block:
        #  - The name, f, of the function which is best at any latency
        #  - The latency, l, for which f is best
        #  - The log p-value, p, for f at l

        # sparse.COO doesn't implement argmin, so we have to do it in a few steps

        data: DataArray = self._data[block_name].copy()
        densify_data_block(data)

        # (channel, function) → l, the best latency
        best_latency = data.idxmin(dim=DIM_LATENCY)
        # (channel, function) → log p of best latency for each function
        logp_at_best_latency = data.min(dim=DIM_LATENCY)

        # (channel) → log p of best function (at best latency)
        logp_at_best_function = logp_at_best_latency.min(dim=DIM_FUNCTION)
        # (channel) → f, the best function
        best_function = logp_at_best_latency.idxmin(dim=DIM_FUNCTION)

        # (channel) -> logp of the best function (at the best latency)
        logp_vals = logp_at_best_function.data
        # (channel) -> best function name (at the best latency)
        best_functions = best_function.data
        # (channel) -> latency of the best function
        best_latencies = best_latency.sel({
            # e.g. hexels          -> array([0, ..., 10241])
            self.channel_coord_name: self._channels[block_name],
            #          -> DataArray((hexel) -> function)
            DIM_FUNCTION: best_function
        }).data

        # Cut out channels which have a best log p-val of 0 (i.e. p = 1)
        idxs = logp_vals < 0

        return DataFrame.from_dict({
            self.channel_coord_name: self._channels[block_name][idxs],
            DIM_FUNCTION: best_functions[idxs],
            DIM_LATENCY: best_latencies[idxs],
            "value": logp_vals[idxs],
        })

    def rename(self, functions: dict[str, str]) -> None:
        """
        Renames the functions within an ExpressionSet.

        Supply a dictionary mapping old function names to new function names.

        Raises KeyError if one of the keys in the renaming dictionary is not a function name in the expression set.
        """
        for old, new in functions.items():
            if old not in self.functions:
                raise KeyError(f"{old} is not a function in this expression set")
        for bn, data in self._data.items():
            new_names = []
            for old_name in self._data[bn][DIM_FUNCTION].values:
                if old_name in functions:
                    new_names.append(functions[old_name])
                else:
                    new_names.append(old_name)
            self._data[bn][DIM_FUNCTION] = new_names

    @abstractmethod
    def best_functions(self) -> DataFrame | tuple[DataFrame, ...]:
        """
        Note that channels for which the best p-value is 1 will be omitted.
        """
        pass


class HexelExpressionSet(ExpressionSet):
    """
    Brain data associated with expression of a single function in hexel space.
    Includes lh, rh, flipped, non-flipped.
    Data is log10 p-values
    """

    def __init__(self,
                 functions: str | Sequence[str],
                 # Metadata
                 hexels_lh: Sequence[Hexel],
                 hexels_rh: Sequence[Hexel],
                 latencies: Sequence[Latency],
                 # log p-values
                 # In general, we will combine flipped and non-flipped versions
                 data_lh: _InputDataArray | Sequence[_InputDataArray],
                 data_rh: _InputDataArray | Sequence[_InputDataArray],
                 ):

        super().__init__(
            functions=functions,
            latencies=latencies,
            data_blocks={
                BLOCK_LEFT: data_lh,
                BLOCK_RIGHT: data_rh,
            },
            channel_coord_name=DIM_HEXEL,
            channel_coord_dtype=HexelDType,
            channel_coord_values={
                BLOCK_LEFT: hexels_lh,
                BLOCK_RIGHT: hexels_rh
            },
        )

    @property
    def hexels_left(self) -> NDArray[HexelDType]:
        """Hexels, canonical ID."""
        return self._channels[BLOCK_LEFT]

    @property
    def hexels_right(self) -> NDArray[HexelDType]:
        """Hexels, canonical ID."""
        return self._channels[BLOCK_RIGHT]

    @property
    def left(self) -> DataArray:
        """Left-hemisphere data."""
        return self._data[BLOCK_LEFT]

    @property
    def right(self) -> DataArray:
        """Right-hemisphere data."""
        return self._data[BLOCK_RIGHT]

    def __getitem__(self, functions: str | Sequence[str]) -> HexelExpressionSet:
        """
        Select data for specified function(s) only.
        Use a function name or list/array of function names
        """
        # Allow indexing by a single function
        if isinstance(functions, str):
            functions = [functions]
        for f in functions:
            if f not in self.functions:
                raise KeyError(f)
        return HexelExpressionSet(
            functions=functions,
            hexels_lh=self.hexels_left,
            hexels_rh=self.hexels_right,
            latencies=self.latencies,
            data_lh=[self._data[BLOCK_LEFT].sel({DIM_FUNCTION: function}).data for function in functions],
            data_rh=[self._data[BLOCK_RIGHT].sel({DIM_FUNCTION: function}).data for function in functions],
        )

    def __copy__(self):
        return HexelExpressionSet(
            functions=self.functions.copy(),
            hexels_lh=self.hexels_left.copy(),
            hexels_rh=self.hexels_right.copy(),
            latencies=self.latencies.copy(),
            data_lh=self._data[BLOCK_LEFT].values.copy(),
            data_rh=self._data[BLOCK_RIGHT].values.copy(),
        )

    def __add__(self, other: HexelExpressionSet) -> HexelExpressionSet:
        assert array_equal(self.hexels_left, other.hexels_left), "Hexels mismatch (left)"
        assert array_equal(self.hexels_right, other.hexels_right), "Hexels mismatch (right)"
        assert array_equal(self.latencies, other.latencies), "Latencies mismatch"
        # constructor expects a sequence of function names and sequences of 2d matrices
        functions = []
        data_lh = []
        data_rh = []
        for expr_set in [self, other]:
            for i, function in enumerate(expr_set.functions):
                functions.append(function)
                data_lh.append(expr_set._data[BLOCK_LEFT].data[:, :, i])
                data_rh.append(expr_set._data[BLOCK_RIGHT].data[:, :, i])
        return HexelExpressionSet(
            functions=functions,
            hexels_lh=self.hexels_left, hexels_rh=self.hexels_right,
            latencies=self.latencies,
            data_lh=data_lh, data_rh=data_rh,
        )

    def __eq__(self, other: HexelExpressionSet) -> bool:
        if not super().__eq__(other):
            return False
        if not COO(self.left.data == other.left.data).all():
            return False
        if not COO(self.right.data == other.right.data).all():
            return False
        return True

    def best_functions(self) -> Tuple[DataFrame, DataFrame]:
        """
        Return a pair of DataFrames (left, right), containing:
        for each hexel, the best function and latency for that hexel, and the associated log p-value

        Note that channels for which the best p-value is 1 will be omitted.
        """
        return (
            super()._best_functions_for_block(BLOCK_LEFT),
            super()._best_functions_for_block(BLOCK_RIGHT),
        )


class SensorExpressionSet(ExpressionSet):
    """
    Brain data associated with the expression of a single function in sensor space.
    Includes left hemisphere (lh), right hemisphere (rh), flipped, and non-flipped data.
    Data is represented as log10 p-values.
    """

    def __init__(self,
                 functions: str | Sequence[str],
                 # Metadata
                 sensors: Sequence[Sensor],
                 latencies: Sequence[Latency],
                 # log p-values
                 # In general, we will combine flipped and non-flipped versions
                 data: _InputDataArray | Sequence[_InputDataArray],
                 ):
        """
        Initialize the SensorExpressionSet with function names, sensor metadata, latency information, and log p-value data.

        Args:
            functions (str | Sequence[str]): The names of the functions being evaluated.
            sensors (Sequence[Sensor]): Metadata about the sensors used in the study.
            latencies (Sequence[Latency]): Latency information corresponding to the data.
            data (_InputDataArray | Sequence[_InputDataArray]): Log p-values representing the data.
        """
        super().__init__(
            functions=functions,
            latencies=latencies,
            data_blocks={
                BLOCK_SCALP: data
            },
            channel_coord_name=DIM_SENSOR,
            channel_coord_dtype=SensorDType,
            channel_coord_values={BLOCK_SCALP: sensors},
        )

    @property
    def sensors(self) -> NDArray[SensorDType]:
        """
        Get the sensor metadata.

        Returns:
            NDArray[SensorDType]: Array of sensor metadata.
        """
        return self._channels[BLOCK_SCALP]

    @property
    def scalp(self) -> DataArray:
        """
        Get the left-hemisphere data.
        """
        return self._data[BLOCK_SCALP]

    def __eq__(self, other: SensorExpressionSet) -> bool:
        if not super().__eq__(other):
            return False
        if not COO(self.scalp.data == other.scalp.data).all():
            return False
        return True

    def __copy__(self):
        return SensorExpressionSet(
            functions=self.functions.copy(),
            sensors=self.sensors.copy(),
            latencies=self.latencies.copy(),
            data=self._data[BLOCK_SCALP].values.copy(),
        )

    def __add__(self, other: SensorExpressionSet) -> SensorExpressionSet:
        assert array_equal(self.sensors, other.sensors), "Sensors mismatch"
        assert array_equal(self.latencies, other.latencies), "Latencies mismatch"
        # constructor expects a sequence of function names and sequences of 2d matrices
        functions = []
        data = []
        for expr_set in [self, other]:
            for i, function in enumerate(expr_set.functions):
                functions.append(function)
                data.append(expr_set._data[BLOCK_SCALP].data[:, :, i])
        return SensorExpressionSet(
            functions=functions,
            sensors=self.sensors, latencies=self.latencies,
            data=data,
        )

    def __getitem__(self, functions: str | Sequence[str]) -> SensorExpressionSet:
        """
        Select data for specified function(s) only.
        Use a function name or list/array of function names
        """
        # Allow indexing by a single function
        if isinstance(functions, str):
            functions = [functions]
        for f in functions:
            if f not in self.functions:
                raise KeyError(f)
        return SensorExpressionSet(
            functions=functions,
            sensors=self.sensors,
            latencies=self.latencies,
            data=[self._data[BLOCK_SCALP].sel({DIM_FUNCTION: function}).data for function in functions],
        )

    def best_functions(self) -> DataFrame:
        """
        Return a DataFrame containing:
        for each sensor, the best function and latency for that sensor, and the associated log p-value

        Note that channels for which the best p-value is 1 will be omitted.
        """
        return super()._best_functions_for_block(BLOCK_SCALP)



from numpy import log10
from numpy.typing import ArrayLike


# The canonical log base
log_base = 10


def p_to_logp(arraylike: ArrayLike) -> ArrayLike:
    """The one-stop-shop for converting from p-values to log p-values."""
    return log10(arraylike)


def logp_to_p(arraylike: ArrayLike) -> ArrayLike:
    """The one-stop-shop for converting from log p-values to p-values."""
    return float(10) ** arraylike
